package dev.pfaff.altargui;

import dev.pfaff.altargui.input.InputKey;
import dev.pfaff.altargui.input.InputMod;
import dev.pfaff.altargui.inspector.DebugLayer;
import dev.pfaff.altargui.inspector.DebugPropertyValue;
import dev.pfaff.altargui.inspector.DetailedRawObjectValue;
import dev.pfaff.altargui.selector.Selector;
import dev.pfaff.altargui.selector.SyntaxError;
import dev.pfaff.altargui.types.Color;
import dev.pfaff.altargui.util.ListMultiChildBuilder;
import dev.pfaff.altargui.util.RepeatMultiChildBuilder;
import dev.pfaff.altargui.util.StringCursor;
import dev.pfaff.altargui.util.listenable.BooleanSignal;
import dev.pfaff.altargui.util.listenable.EventListenable;
import dev.pfaff.altargui.util.listenable.SlotValue;
import dev.pfaff.altargui.widgets.*;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;
import it.unimi.dsi.fastutil.shorts.ShortArrayList;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jspecify.annotations.NonNull;

import java.util.Arrays;
import java.util.BitSet;
import java.util.List;
import java.util.function.Predicate;
import java.util.function.Supplier;

import static dev.pfaff.altargui.CanvasImpl.DEBUG_LAYERS;
import static dev.pfaff.altargui.Key.identityKey;
import static dev.pfaff.altargui.util.ListBuilder.buildWidgetList;
import static dev.pfaff.altargui.util.ListBuilder.widgetListBuilder;
import static dev.pfaff.altargui.widgets.MultiChildWidget.*;
import static dev.pfaff.altargui.widgets.TextWidget.DEFAULT_SCALED_FONT_SIZE;
import static java.util.FormatProcessor.FMT;

public final class Inspector implements SimpleBuiltWidget {
	private static final int INDENT = 6;
	private static final short[] EMPTY_SHORT_ARRAY = new short[0];
	private static final Widget TREE_LINE = RectWidget.colored(Color.fromArgb(0xff_aaaaaa))
														.withAbsoluteWidth(1)
														.withHorizontalMargin(2, INDENT-3-1);

	public static final float OBJECT_INSPECTOR_Y_SPACING = 1.5f;

//	public static final float FONT_SIZE = DEFAULT_SCALED_FONT_SIZE / 2f;
	public static final float FONT_SIZE = DEFAULT_SCALED_FONT_SIZE;
	private static final float MIN_INSPECTOR_WIDTH = 64f;
	private static final float MIN_INSPECTOR_HEIGHT = 32f;
	private static final Widget DRAG_BORDER_RECT = RectWidget.colored(Color.fromRgb(0x202020));

	public static TextWidget withInspectorFontSize(TextWidget widget) {
		return widget.withScaledFontSize(FONT_SIZE);
	}

	private static final Widget PROPERTY_LAYOUT_X = withInspectorFontSize(TextWidget.of("x"));
	private static final Widget PROPERTY_LAYOUT_Y = withInspectorFontSize(TextWidget.of("y"));
	private static final Widget PROPERTY_LAYOUT_ALLOC_W = withInspectorFontSize(TextWidget.of("allocW"));
	private static final Widget PROPERTY_LAYOUT_ALLOC_H = withInspectorFontSize(TextWidget.of("allocH"));
	private static final Widget PROPERTY_LAYOUT_COMP_W = withInspectorFontSize(TextWidget.of("compW"));
	private static final Widget PROPERTY_LAYOUT_COMP_H = withInspectorFontSize(TextWidget.of("compH"));
	public static final Widget CLICK_TO_INSPECT_TEXT = withInspectorFontSize(TextWidget.of("Click to inspect"));
	public static final Widget NAME_VALUE_SEPARATOR = withInspectorFontSize(TextWidget.of(": "));
	public static final Widget CLICK_TO_INSPECT_FIELD_TEXT = withInspectorFontSize(TextWidget.of("^ "));

	private final GuiContext context;
	private final WidgetInstanceList orderedElements;
	private final ShortArrayList elementDepths;
	private final BitSet activeDebugLayers;
	private final EventListenable activeDebugLayersUpdate = new EventListenable();
	private final WidgetInstanceList filteredElementsBuffer = new WidgetInstanceList(512);
	private short[] filteredElementDepthsBuffer = EMPTY_SHORT_ARRAY;
	private final ObjectArrayList<InspectingTarget> inspectingObjectStack = new ObjectArrayList<>();
	private final EventListenable inspectionTargetUpdate = new EventListenable();
	private final SlotValue<@Nullable WidgetInstance<?>> hoveredElement = new SlotValue<>(null);
	private final BooleanSignal isPickingElement = new BooleanSignal(false);
	private final SlotValue<Widget> elementSelectorError = new SlotValue<>(Widgets.EMPTY);

	// set in GuiContext
	int inspectableCount = 0;

	private final SlotValue<InspectionTargetGroup> inspectionTargetGroup = new SlotValue<>(InspectionTargetGroup.Elements);
	private long selectedElement = -1;

	private final Widget objectInspectorBackButton = ButtonWidget.inlineStyle(withInspectorFontSize(TextWidget.of("< Back")),
					() -> {
						this.inspectingObjectStack.pop();
						this.inspectionTargetUpdate.notifyListeners();
					})
																 .withMarginEnd(6);

	public final SlotValue<Boolean> isOpen = new SlotValue<>(false);
	private final CanvasImpl canvas;
	private final Widget pickElementWidget;

	private sealed interface InspectingTarget {
		Widget widget();
	}

	private record InspectingObject(Widget widget) implements InspectingTarget {}

	private record InspectingElement(WidgetInstance<?> element, Widget widget) implements InspectingTarget {}

	Inspector(GuiContext context, WidgetInstanceList orderedElements, ShortArrayList elementDepths, BitSet layers, CanvasImpl canvas) {
		this.context = context;
		this.orderedElements = orderedElements;
		this.elementDepths = elementDepths;
		this.activeDebugLayers = layers;
		this.canvas = canvas;

		this.pickElementWidget = new MouseOverWithCoordsBuilder((hovered, mouseX, mouseY) -> {
			return new FrameBuilderWidget(_ -> {
				var consumer = new Predicate<WidgetInstance<?>>() {
					@Nullable
					private WidgetInstance<?> element;

					@Override
					public boolean test(WidgetInstance<?> instance) {
						var offX = mouseX - instance.allocatedX();
						var offY = mouseY - instance.allocatedY();
						if (offX >= 0.0 &&
							   offY >= 0.0 &&
							   offX < instance.computedWidth() &&
							   offY < instance.computedHeight()) {
							this.element = instance;
							return true;
						}
						return false;
					}
				};
				this.orderedElements.iterateDown(consumer);
				if (consumer.element != null) {
					return buildElementHighlightOnce(consumer.element, new ButtonWidget(ELEMENT_HIGHLIGHT_RECT, () -> {
						this.clearInspectingObjectStack();
						this.pushInspectionTargetElement(consumer.element);
						this.isPickingElement.clear();
					}));
				} else {
					return Widgets.EMPTY;
				}
			}).expand();
		});
	}

	private record DebugLayerToggleWidget(Inspector inspector, DebugLayer layer) implements SimpleBuiltWidget {
		@Override
		public Widget build() {
			var inspector = this.inspector;
			return ButtonWidget.inlineStyle(row(List.of(
					new ListenableBuilderWidget(inspector.activeDebugLayersUpdate,
							() -> Widgets.checkbox(inspector.activeDebugLayers.get(layer.id())))
							.withMarginEnd(3),
					withInspectorFontSize(TextWidget.of(layer.name))
							.withMarginTop(0.5f)
			))
					.withMargin(2), () -> {
				inspector.activeDebugLayers.flip(layer.id());
				inspector.activeDebugLayersUpdate.notifyListeners();
			});
		}
	}

	private static Widget indentedCol(MultiChildWidget.Builder builder) {
		return col(ctx -> {
			builder.build((callSite, child) -> {
				ctx.layout(callSite, child.withVerticalMargin(1f));
			});
		}).withMarginStart(INDENT);
	}

	public static Widget labeledCol(String label, MultiChildWidget.Builder builder) {
		return col(List.of(withInspectorFontSize(TextWidget.of(label)).withUnderline(true),
						   indentedCol(builder)
		)).withMarginTop(OBJECT_INSPECTOR_Y_SPACING);
	}

	public static Widget labeledCol(String label, List<? extends Widget> children) {
		return labeledCol(label, ListMultiChildBuilder.newUntrusted(children));
	}

	public static Widget buildKeyValueEntry(Widget name, Widget widget) {
		return row(List.of(name, NAME_VALUE_SEPARATOR, widget))
				.withSymmetricMargin(0, 0.5f);
//		return row(mounter -> {
//			mounter.layout(() -> {}, name);
//			mounter.layout(() -> {}, NAME_VALUE_SEPARATOR);
//			mounter.layout(() -> {}, widget);
//		}).withSymmetricMargin(0, 0.5f);
	}

	public Widget buildValuePreview(Widget preview, @Nullable Supplier<Widget> detailed) {
		if (detailed != null) {
			return ButtonWidget.inlineStyle(preview, () -> {
				this.inspectingObjectStack.push(new InspectingObject(makeInspectionTargetWidget(preview, detailed.get())));
				this.inspectionTargetUpdate.notifyListeners();
			}).withTooltip(CLICK_TO_INSPECT_TEXT);
		} else {
			return preview;
		}
	}

	public Widget buildValuePreview(DebugPropertyValue value) {
		return buildValuePreview(value.display(), value instanceof DetailedRawObjectValue<?> detailed
				? () -> detailed.displayDetailed(this) : null);
	}

	public Widget buildKeyValueEntry(Widget name, DebugPropertyValue value) {
		return buildKeyValueEntry(name, buildValuePreview(value));
	}

	public void pushInspectionTargetObject(Object value) {
		this.inspectingObjectStack.push(new InspectingObject(makeInspectionTargetWidget((DetailedRawObjectValue<?>) DebugPropertyValue.of(value))));
		this.inspectionTargetUpdate.notifyListeners();
	}

	public void pushInspectionTargetElement(WidgetInstance<?> element) {
		this.selectedElement = element.key;
		this.inspectingObjectStack.push(new InspectingElement(element, makeInspectionTargetWidget((DetailedRawObjectValue<?>) DebugPropertyValue.of(element))));
		this.inspectionTargetUpdate.notifyListeners();
	}

	private Widget makeInspectionTargetWidget(DetailedRawObjectValue<?> value) {
		return this.makeInspectionTargetWidget(value.display(), value.displayDetailed(this));
	}

	public Widget makeInspectionTargetWidget(@NotNull Widget preview, @NotNull Widget detailed) {
		int inspectingObjectStackSize = this.inspectingObjectStack.size();
		return col(List.of(
			row(widgetListBuilder(2)
						.addIf(inspectingObjectStackSize >= 1, this.objectInspectorBackButton)
						.add(preview)
						.build())
					.withMarginBottom(2f),
			new ScrollWidget(detailed)
		)).withMargin(3);
	}

	private static String getNestedName(Class<?> clazz) {
		String name = clazz.getSimpleName();
		var host = clazz.getNestHost();
		if (host == null || host == clazz) return name;
		return getNestedName(host) + "." + name;
	}

	private record ElementItemWidget(Inspector inspector, WidgetInstance<?> element, short depth) implements SimpleBuiltWidget {
		@Override
		public Widget build() {
			var inspector = this.inspector;
			var element = this.element;
			var depth = this.depth;
			var labelStr = FMT."\{getNestedName(element.widgetClass)} [%016x\{element.key}]";
			var labelText = withInspectorFontSize(TextWidget.of(labelStr).withBold(true));
			var item = new MouseMoveTarget(ButtonWidget.inlineStyle(labelText.withMargin(0, 0.5f, 0, 0).withMargin(2), () -> {
				int n = inspector.inspectingObjectStack.size();
				inspector.clearInspectingObjectStack();
				if (inspector.selectedElement == element.key) {
					if (n == 1) {
						inspector.selectedElement = -1;
						return;
					}
				}
				inspector.pushInspectionTargetElement(element);
			}), hovered -> {
				if (hovered) {
					inspector.hoveredElement.write(element);
				} else {
					// this nullability warning (value must be non-null) is wrong
					inspector.hoveredElement.atomicCompareAndSet(element, null);
				}
				return false;
			});
			return row(builder -> {
				builder.layout(() -> {}, row(new RepeatMultiChildBuilder(TREE_LINE, depth))
						.withMultiChildAttributes()
						.tightenHeight());
				builder.layout(() -> {}, item);
			});
		}

		@Override
		public boolean equals(Object obj) {
			if (obj == this) return true;
			if (!(obj instanceof ElementItemWidget(var inspector, var element, var depth))) return false;
			return this.inspector == inspector && this.element == element && this.depth == depth;
		}
	}

	private void clearInspectingObjectStack() {
		if (this.inspectingObjectStack.isEmpty()) return;
		this.inspectingObjectStack.clear();
		this.inspectionTargetUpdate.notifyListeners();
	}

	private Widget buildInspectorTargetSelector() {
		return ValueListenableBuilderWidget.builder(this.inspectionTargetGroup, group -> {
			return col(buildWidgetList(2, builder -> {
				builder.add(row(List.of(
						ButtonWidget.inlineStyle(InspectionTargetGroup.Elements.label, () -> {
							if (group != InspectionTargetGroup.Elements) {
								this.inspectionTargetGroup.write(InspectionTargetGroup.Elements);
								this.selectedElement = -1;
								clearInspectingObjectStack();
							}
						}).withMargin(3),
//						ButtonWidget.inlineStyle(InspectionTargetGroup.Game.label, () -> {
//							if (group != InspectionTargetGroup.Game) {
//								this.inspectionTargetGroup.value(InspectionTargetGroup.Game);
//								clearInspectingObjectStack();
//								pushInspectionTargetObject(client);
//							}
//						}).withMargin(3),
						ButtonWidget.inlineStyle(InspectionTargetGroup.GuiContext.label, () -> {
							if (group != InspectionTargetGroup.GuiContext) {
								this.inspectionTargetGroup.write(InspectionTargetGroup.GuiContext);
								clearInspectingObjectStack();
								pushInspectionTargetObject(context);
							}
						}).withMargin(3)
				)));

				switch (group) {
					case Elements -> {
						builder.add(new FrameBuilderWidget(_ -> {
							int inspectableCount = this.inspectableCount;
							var a = filteredElementsBuffer.ensureCapacityAndReturnArray(inspectableCount);
							if (filteredElementDepthsBuffer.length != a.length) {
								filteredElementDepthsBuffer = new short[a.length];
							}
							System.arraycopy(orderedElements.elements(), 0, a, 0, inspectableCount);
							elementDepths.getElements(0, filteredElementDepthsBuffer, 0, inspectableCount);
							filteredElementsBuffer.setSizeAssumeCapacityAndInitialized(inspectableCount);
							Arrays.fill(a, inspectableCount, a.length, null);

							return col(List.of(
									new TextFieldWidget().withOnChanged(s -> {
										if (s.isBlank()) {
											elementSelectorError.write(Widgets.EMPTY);
											return;
										}
										Selector selector;
										try {
											selector = Selector.parse(new StringCursor(s));
										} catch (SyntaxError e) {
											elementSelectorError.write(TextWidget.of(e.toString()));
											return;
										}
										if (!filteredElementsBuffer.isEmpty()) {
											var element = selector.selectFirst(filteredElementsBuffer.getFirst());
											clearInspectingObjectStack();
											if (element != null) {
												pushInspectionTargetElement(element);
												elementSelectorError.write(Widgets.EMPTY);
											} else {
												selectedElement = -1;
												elementSelectorError.write(TextWidget.of("No matching element"));
											}
										} else {
											elementSelectorError.write(TextWidget.of("No elements to select from"));
										}
									}).withTooltip(new ValueListenableBuilderWidget(elementSelectorError)),
									new FiniteScrollListWidget(inspectableCount,
											i -> identityKey(a[i]),
											i -> new ElementItemWidget(this,
													a[i],
													this.filteredElementDepthsBuffer[i]))));
						}));
					}
					case GuiContext -> {
					}
				}
			}));
		});
	}

	private static final float ELEMENT_HIGHLIGHT_THICKNESS = 2f;
	private static final Widget ELEMENT_HIGHLIGHT_RECT = RectWidget.colored(Color.fromArgb(0x55_0000ff))
			.withBorder(Color.fromArgb(0xFF_0000FF), ELEMENT_HIGHLIGHT_THICKNESS);

	private static Widget buildElementHighlightOnce(@NonNull WidgetInstance<?> element, Widget widget) {
		return widget
				.withAbsoluteSize(element.computedWidth() + ELEMENT_HIGHLIGHT_THICKNESS * 2,
						element.computedHeight() + ELEMENT_HIGHLIGHT_THICKNESS * 2)
				.atAbsolutePosition(element.allocatedX() - ELEMENT_HIGHLIGHT_THICKNESS,
						element.allocatedY() - ELEMENT_HIGHLIGHT_THICKNESS);
	}

	private static Widget buildElementHighlight(@Nullable WidgetInstance<?> element) {
		if (element == null) return Widgets.EMPTY;
		return new FrameBuilderWidget(_ -> buildElementHighlightOnce(element, ELEMENT_HIGHLIGHT_RECT));
	}

	private static  <T> @Nullable T topOrNull(ObjectArrayList<T> list) {
		return !list.isEmpty() ? list.top() : null;
	}

	public Widget build() {
		var inspectingObjectStack = this.inspectingObjectStack;

		var windowWidth = new SlotValue<>(400f);
		var windowHeight = new SlotValue<>(175f);

		var dragBorderThickness = 3f;

		var windowContent = new ClipWidget(row(List.of(
			col(List.of(
				new ListenableBuilderWidget(this.isPickingElement, () -> {
					if (this.isPickingElement.get()) {
						return ButtonWidget.inlineStyle(withInspectorFontSize(TextWidget.of("Unpick Element")),
								this.isPickingElement::clear);
					} else {
						return ButtonWidget.inlineStyle(withInspectorFontSize(TextWidget.of("Pick Element")),
								this.isPickingElement::set);
					}
				}).withMargin(3),
				ButtonWidget.inlineStyle(withInspectorFontSize(TextWidget.of("Dump Texture")), () -> {
					this.canvas.texAlloc().dump();
				}).withMargin(3),
				new FiniteScrollListWidget(DEBUG_LAYERS.size(),
						i -> new DebugLayerToggleWidget(this, DEBUG_LAYERS.get(i)))
			)),
			this.buildInspectorTargetSelector(),
			new ListenableBuilderWidget(this.inspectionTargetUpdate, () -> {
				var target = topOrNull(inspectingObjectStack);
				return target != null ? target.widget() : Widgets.EMPTY;
			})
		))).expand().withBackground(RectWidget.GENERIC_HEAVY_HUD_BACKGROUND);

		var windowContentWithHeight = ValueListenableBuilderWidget.builder(windowHeight, height -> windowContent.withAbsoluteHeight(height - dragBorderThickness));
		var windowContentWithWidth = ValueListenableBuilderWidget.builder(windowWidth, width -> windowContentWithHeight.withAbsoluteWidth(width - dragBorderThickness));

		var heightResizeHandle = new DragTarget(DRAG_BORDER_RECT.withAbsoluteHeight(dragBorderThickness),
			(dragging, _, _, _, deltaY) -> {
				if (dragging && deltaY != 0) {
					windowHeight.write(Math.max(windowHeight.read() + deltaY,
							MIN_INSPECTOR_HEIGHT));
				}
			});
		var widthResizeHandle = new DragTarget(DRAG_BORDER_RECT.withAbsoluteWidth(dragBorderThickness),
			(dragging, _, _, deltaX, _) -> {
				if (dragging && deltaX != 0) {
					windowWidth.write(Math.max(windowWidth.read() + deltaX, MIN_INSPECTOR_WIDTH));
				}
			});

		Widget whenOpen = stack(List.of(
			new ListenableBuilderWidget(this.inspectionTargetUpdate, () -> {
				return topOrNull(inspectingObjectStack) instanceof InspectingElement elTarget
						? buildElementHighlight(elTarget.element)
						: Widgets.EMPTY;
			}),
			new ValueListenableBuilderWidget(this.hoveredElement.map(Inspector::buildElementHighlight)),
			new ListenableBuilderWidget(this.isPickingElement, () -> {
				if (this.isPickingElement.get()) {
					return this.pickElementWidget;
				} else {
					return Widgets.EMPTY;
				}
			}),
			new DraggableWidget(col(List.of(
				row(List.of(
					windowContentWithWidth,
					ValueListenableBuilderWidget.builder(windowHeight, height -> {
						return widthResizeHandle.withAbsoluteHeight(height - dragBorderThickness);
					})
				)),
				row(List.of(
					ValueListenableBuilderWidget.builder(windowWidth, width -> {
						return heightResizeHandle.withAbsoluteWidth(width - dragBorderThickness);
					}),
					new DragTarget(DRAG_BORDER_RECT.withAbsoluteSize(dragBorderThickness, dragBorderThickness),
							(dragging, _, _, deltaX, deltaY) -> {
								if (dragging && (deltaX != 0 || deltaY != 0)) {
									windowWidth.write(Math.max(windowWidth.read() + deltaX, MIN_INSPECTOR_WIDTH));
									windowHeight.write(Math.max(windowHeight.read() + deltaY,
											MIN_INSPECTOR_HEIGHT));
								}
							})
				))
			)),16, 16)
		));
		return stack(List.of(
				ValueListenableBuilderWidget.builder(this.isOpen, isOpen -> isOpen ? whenOpen : Widgets.EMPTY),
				new HotKeyWidget((keyCode, scanCode, modifiers) -> {
					if (keyCode == InputKey.LEFT_BRACKET && modifiers.has(InputMod.ALT)) {
						this.canvas.nextDebugLayer(false);
						return true;
					} else if (keyCode == InputKey.RIGHT_BRACKET && modifiers.has(InputMod.ALT)) {
						this.canvas.nextDebugLayer(true);
						return true;
					} else if (keyCode == InputKey.I && modifiers.has(InputMod.CTRL) && modifiers.has(InputMod.SHIFT)) {
						this.isOpen.write(!this.isOpen.read());
						return true;
					}
					return false;
				})
		));
	}

	private enum InspectionTargetGroup {
		Elements,
		GuiContext;

		public final Widget label = withInspectorFontSize(TextWidget.of(this.name()));
	}
}
