package dev.pfaff.altargui.io;

import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.SegmentAllocator;

@FunctionalInterface
public interface Reallocator {
	/**
	 * Reallocates the given segment to be the specified size. Upon returning, the original segment may be invalid.
	 *
	 * @param segment       the segment to reallocate. This must be either {@link MemorySegment#NULL} or the return
	 *                      value
	 *                      of a prior invocation to this method on the same {@link Reallocator} instance.
	 * @param byteSize      the size (in bytes) of the reallocated segment
	 * @param byteAlignment the alignment (in bytes) of the reallocated segment
	 * @return the new segment
	 */
	MemorySegment reallocate(MemorySegment segment, long byteSize, long byteAlignment);

	/**
	 * Like {@link #reallocate(MemorySegment, long, long)} but the {@code byteAlignment} is taken from
	 * {@link MemorySegment#maxByteAlignment() segment.maxByteAlignment()}.
	 */
	default MemorySegment reallocate(MemorySegment segment, long byteSize) {
		return reallocate(segment, byteSize, segment.maxByteAlignment());
	}

	static Reallocator ofAuto() {
		return (segment, byteSize, byteAlignment) -> Arena.ofAuto().allocate(byteSize, byteAlignment).copyFrom(segment);
	}

	static Reallocator usingArena(Arena arena) {
		return (segment, byteSize, byteAlignment) -> arena.allocate(byteSize, byteAlignment).copyFrom(segment);
	}

	static Reallocator usingAllocator(SegmentAllocator allocator) {
		return (segment, byteSize, byteAlignment) -> allocator.allocate(byteSize, byteAlignment).copyFrom(segment);
	}
}
