package dev.pfaff.altargui.libimage;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.util.MethodHandleUtil;
import dev.pfaff.log4truth.Logger;
import dev.pfaff.log4truth.format.Format;

import java.lang.foreign.*;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.VarHandle;
import java.util.List;

import static dev.pfaff.altargui.util.MethodHandleUtil.withoutBase;
import static dev.pfaff.glass.Interop.*;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static java.lang.foreign.MemoryLayout.PathElement.groupElement;
import static java.lang.foreign.MemoryLayout.structLayout;
import static java.lang.foreign.MemoryLayout.unionLayout;
import static java.lang.foreign.ValueLayout.*;
import static java.lang.invoke.MethodHandles.*;
import static java.lang.invoke.MethodType.methodType;

public final class LibImage {
	private static final Linker LINKER = Linker.nativeLinker();
	private static final SymbolLookup LOOKUP = GuiGlobals.HOST.loadLibImage();

	private static final ScopedValue<Arena> ALLOC_ARENA = ScopedValue.newInstance();

	private static final ValueLayout ML_ColorType = JAVA_BYTE.withName(ColorType.class.getSimpleName());
	private static final ValueLayout ML_ImageFormat = JAVA_BYTE.withName(ImageFormat.class.getSimpleName());
	private static final ValueLayout ML_EncoderProfile = JAVA_BYTE.withName(EncoderProfile.class.getSimpleName());
	private static final StructLayout ML_ImageInfo = structLayout(JAVA_INT.withName("width"),
			JAVA_INT.withName("height"),
			ML_ColorType.withName("color_type"),
			MemoryLayout.paddingLayout(Integer.BYTES - Byte.BYTES));

	private static final StructLayout ML_OkLoadedImage = structLayout(
			ADDRESS.withTargetLayout(JAVA_BYTE).withName("ptr"),
			ADDRESS_INT.withName("len"),
			ML_ImageInfo.withName("info"),
			MemoryLayout.paddingLayout(Integer.BYTES)
	);
	private static final StructLayout ML_ErrLoadedImage = structLayout(
			ADDRESS.withName("marker")
	);
	private static final UnionLayout ML_LoadedImage = unionLayout(
			ML_OkLoadedImage.withName("ok"),
			ML_ErrLoadedImage.withName("err")
	);

	private static final StructLayout ML_OkEncodedImage = structLayout(
			ADDRESS.withTargetLayout(JAVA_BYTE).withName("ptr"),
			ADDRESS_INT.withName("len")
	);
	private static final StructLayout ML_ErrEncodedImage = structLayout(
			ADDRESS.withName("marker")
	);
	private static final UnionLayout ML_EncodedImage = unionLayout(
			ML_OkEncodedImage.withName("ok"),
			ML_ErrEncodedImage.withName("err")
	);

	public static final PathElement PE_ok = groupElement("ok");
	public static final PathElement PE_info = groupElement("info");

	private static final VarHandle VH_LoadedImage_ok_ptr =
			withoutBase(ML_LoadedImage.varHandle(PE_ok, groupElement("ptr")).withInvokeExactBehavior());
	private static final VarHandle VH_LoadedImage_ok_len = filterValue(
			withoutBase(ML_LoadedImage.varHandle(PE_ok, groupElement("len")).withInvokeExactBehavior()),
			MH_cast_long_to_usize,
			MH_cast_usize_to_long
	);
	private static final VarHandle VH_LoadedImage_ok_info_width =
			withoutBase(ML_LoadedImage.varHandle(PE_ok, PE_info, groupElement("width")).withInvokeExactBehavior());
	private static final VarHandle VH_LoadedImage_ok_info_height =
			withoutBase(ML_LoadedImage.varHandle(PE_ok, PE_info, groupElement("height")).withInvokeExactBehavior());
	private static final VarHandle VH_LoadedImage_ok_info_color_type =
			withoutBase(ML_LoadedImage.varHandle(PE_ok, PE_info, groupElement("color_type")).withInvokeExactBehavior());

	private static final VarHandle VH_EncodedImage_ok_ptr =
			withoutBase(ML_EncodedImage.varHandle(PE_ok, groupElement("ptr")).withInvokeExactBehavior());
	private static final VarHandle VH_EncodedImage_ok_len = filterValue(
			withoutBase(ML_EncodedImage.varHandle(PE_ok, groupElement("len")).withInvokeExactBehavior()),
			MH_cast_long_to_usize,
			MH_cast_usize_to_long
	);

	private static final VarHandle VH_ImageInfo_width =
			withoutBase(ML_ImageInfo.varHandle(groupElement("width")).withInvokeExactBehavior());
	private static final VarHandle VH_ImageInfo_height =
			withoutBase(ML_ImageInfo.varHandle(groupElement("height")).withInvokeExactBehavior());
	private static final VarHandle VH_ImageInfo_color_type =
			withoutBase(ML_ImageInfo.varHandle(groupElement("color_type")).withInvokeExactBehavior());

	private static final MethodHandle MH_load_image = filterArguments(LINKER.downcallHandle(LOOKUP.find(
							"load_image").get(),
					FunctionDescriptor.of(ML_LoadedImage,
							ADDRESS.withTargetLayout(JAVA_BYTE).withName("input_ptr"),
							ADDRESS_INT.withName("input_len"),
							ADDRESS.withName("alloc"))),
			2,
			MH_cast_long_to_usize);
	private static final MethodHandle MH_encode_image = filterArguments(LINKER.downcallHandle(LOOKUP.find(
							"encode_image").get(),
					FunctionDescriptor.of(ML_OkEncodedImage,
							ADDRESS.withTargetLayout(JAVA_BYTE).withName("input_ptr"),
							ADDRESS_INT.withName("input_len"),
							ADDRESS.withName("alloc"),
							ML_ImageInfo,
							ML_ImageFormat,
							ML_EncoderProfile)),
			2,
			MH_cast_long_to_usize);

	private static MemorySegment alloc(long length) {
		return ALLOC_ARENA.get().allocate(length);
	}

	private static final MethodHandle MH_alloc = filterArguments(MethodHandleUtil.findStatic(MethodHandles.lookup(),
			LibImage.class,
			"alloc",
			methodType(MemorySegment.class, long.class)),
			0,
			MH_cast_usize_to_long);
	private static final MemorySegment MS_alloc = Linker.nativeLinker().upcallStub(MH_alloc,
			FunctionDescriptor.of(ADDRESS, ADDRESS_INT),
			Arena.ofAuto());

	public static ImageData loadImage(MemorySegment bytes, Arena pixelsArena) {
		long start = System.nanoTime();
		try (var arena = Arena.ofConfined()) {
			var result = ScopedValue.getWhere(ALLOC_ARENA, pixelsArena, () -> {
				try {
					return (MemorySegment) MH_load_image.invokeExact((SegmentAllocator) arena, bytes, bytes.byteSize(), MS_alloc);
				} catch (Throwable e) {
					throw new RuntimeException(e);
				}
			});
			var ptr = (MemorySegment) VH_LoadedImage_ok_ptr.get(result);
			if (ptr.equals(MemorySegment.NULL)) {
				throw new RuntimeException("Failed to load png info");
			}
			var len = (long) VH_LoadedImage_ok_len.get(result);
			if (len < 0) {
				throw new UnsupportedOperationException(STR."Unsupported image size \{Long.toUnsignedString(len)} bytes. Must be at most \{Long.MAX_VALUE} bytes.");
			}
			var info = new ImageInfo(
					(int) VH_LoadedImage_ok_info_width.get(result),
					(int) VH_LoadedImage_ok_info_height.get(result),
					ColorType.VALUES.get((byte) VH_LoadedImage_ok_info_color_type.get(result))
			);
			return new ImageData(info, ptr.reinterpret(len));
		} finally {
			long elapsed = System.nanoTime() - start;
			Logger.log(() -> "Completed loadImage in " + Format.STANDARD_DURATION.nanosToString(elapsed),
					List.of(DEBUG));
		}
	}

	/**
	 * @return the encoded image data
	 */
	public static MemorySegment encodeImage(ImageInfo info, MemorySegment pixels, ImageFormat format, EncoderProfile profile, Arena encodedArena) {
		long start = System.nanoTime();
		try (var retArena = Arena.ofConfined()) {
			var result = ScopedValue.getWhere(ALLOC_ARENA, encodedArena, () -> {
				try (var argsArena = Arena.ofConfined()) {
					var infoSegment = argsArena.allocate(ML_ImageInfo);
					VH_ImageInfo_width.set(infoSegment, info.width());
					VH_ImageInfo_height.set(infoSegment, info.height());
					VH_ImageInfo_color_type.set(infoSegment, info.colorType().value());

					return (MemorySegment) MH_encode_image.invokeExact((SegmentAllocator) retArena,
							pixels,
							pixels.byteSize(),
							MS_alloc,
							infoSegment,
							format.value(),
							profile.value());
				} catch (Throwable e) {
					throw new RuntimeException(e);
				}
			});
			var ptr = (MemorySegment) VH_EncodedImage_ok_ptr.get(result);
			if (ptr.equals(MemorySegment.NULL)) {
				throw new RuntimeException("Failed to load png info");
			}
			var len = (long) VH_EncodedImage_ok_len.get(result);
			if (len < 0) {
				throw new UnsupportedOperationException(STR."Unsupported image size \{Long.toUnsignedString(len)} bytes. Must be at most \{Long.MAX_VALUE} bytes.");
			}
			return ptr.reinterpret(len);
		} finally {
			long elapsed = System.nanoTime() - start;
			Logger.log(() -> "Completed encodeImage in " + Format.STANDARD_DURATION.nanosToString(elapsed),
					List.of(DEBUG));
		}
	}
}
