package dev.pfaff.altargui.render;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.glass.opengl.ConstPointer;
import dev.pfaff.glass.opengl.enums.*;
import org.jetbrains.annotations.NotNull;

import java.lang.foreign.MemorySegment;

import static java.lang.foreign.ValueLayout.JAVA_INT;

public final class GLUtil {
	public static @NotNull MemorySegment glMapBuffer(GLBufferTarget target, GLBufferAccess access, long length) {
		var addr = GuiGlobals.GL.glMapBuffer(target, access);
		if (addr == null) throw new NullPointerException();
		var buffer = addr.ms().reinterpret(length);
		return switch (access) {
			case READ_ONLY -> buffer.asReadOnly();
			case WRITE_ONLY, READ_WRITE -> buffer;
		};
	}

	/**
	 * @param access bitset of {@code GL_MAP_*} bits.
	 */
	public static @NotNull MemorySegment glMapBufferRange(GLBufferTarget target, long offset, long length,
														  int access) {
		var addr = GuiGlobals.GL.glMapBufferRange(target, offset, length, access);
		if (addr == null) throw new NullPointerException();
		var buffer = addr.ms().reinterpret(length);
		return (access & GLMapBufferAccessMask.MAP_WRITE_BIT.glEnum()) == 0 ? buffer.asReadOnly() : buffer;
	}

	public static void glUnmapBuffer(GLBufferTarget target) {
		GuiGlobals.GL.glUnmapBuffer(target);
	}

	public static void glBufferData(GLBufferTarget target, long size, GLBufferUsage usage) {
		GuiGlobals.GL.glBufferData(target, size, null, usage);
	}

	public static void glBufferData(GLBufferTarget target, @NotNull MemorySegment data, GLBufferUsage usage) {
		GuiGlobals.GL.glBufferData(target, data.byteSize(), ConstPointer.newNullable(data), usage);
	}

	public static void glBufferStorage(GLBufferStorageTarget target, int size, int flags) {
		GuiGlobals.GL.glBufferStorage(target, size, null, flags);
	}

	public static void glBufferSubData(GLBufferTarget target, long offset, @NotNull MemorySegment data) {
		GuiGlobals.GL.glBufferSubData(target, offset, data.byteSize(), ConstPointer.newNullable(data));
	}

	public static void glCopyBufferSubData(GLCopyBufferSubDataTarget readTarget, GLCopyBufferSubDataTarget writeTarget
			, long readOffset, long writeOffset, long length) {
		GuiGlobals.GL.glCopyBufferSubData(readTarget, writeTarget, readOffset, writeOffset, length);
	}

	public static void glVertexAttribDivisor(int location, int divisor) {
		GuiGlobals.GL.glVertexAttribDivisor(location, divisor);
	}

	public static void glBindVertexArray(int id) {
		GuiGlobals.GL.glBindVertexArray(id);
	}

	public static void glxUnbindVertexArray() {
		GuiGlobals.GL.glBindVertexArray(0);
	}

	public static int glGenBuffers() {
		return GuiGlobals.GL.glGenBuffer();
	}

	public static void glDeleteBuffers(int id) {
		GuiGlobals.GL.glDeleteBuffer(id);
	}

	public static int glGetAttribLocation(int program, String name) {
		return GuiGlobals.GL.glGetAttribLocation(program, name);
	}

	public static void glEnableVertexAttribArray(int id) {
		GuiGlobals.GL.glEnableVertexAttribArray(id);
	}

	/**
	 * @since OpenGL 2.0
	 */
	public static void glVertexAttribPointer(int id, int size, GLVertexAttribPointerType type, boolean normalized,
											 int stride, long pointer) {
		GuiGlobals.GL.glVertexAttribPointer(id,
				size,
				type,
				normalized,
				stride,
				ConstPointer.newNullable(MemorySegment.ofAddress(pointer)));
	}

	/**
	 * @since OpenGL 3.0
	 */
	public static void glVertexAttribIPointer(int id, int size, GLVertexAttribIType type, int stride, long pointer) {
		GuiGlobals.GL.glVertexAttribIPointer(id, size, type, stride, ConstPointer.newNullable(MemorySegment.ofAddress(pointer)));
	}

	/**
	 * @since OpenGL 4.1
	 */
	public static void glVertexAttribLPointer(int id, int size, GLVertexAttribLType type, int stride, long pointer) {
		GuiGlobals.GL.glVertexAttribLPointer(id, size, type, stride, ConstPointer.newNullable(MemorySegment.ofAddress(pointer)));
	}

	public static int glGenVertexArrays() {
		return GuiGlobals.GL.glGenVertexArray();
	}

	public static void glDeleteVertexArrays(int id) {
		GuiGlobals.GL.glDeleteVertexArray(id);
	}

	public static void glDrawArrays(GLPrimitiveType mode, int first, int count) {
		GuiGlobals.GL.glDrawArrays(mode, first, count);
	}

	/**
	 * @param mode          the draw mode.
	 * @param first         the index of the first element in the enabled arrays.
	 * @param count         the number of elements per instance.
	 * @param instanceCount the number of instances.
	 */
	public static void glDrawArraysInstanced(GLPrimitiveType mode, int first, int count, int instanceCount) {
		GuiGlobals.GL.glDrawArraysInstanced(mode, first, count, instanceCount);
	}

	/**
	 * @param mode          the draw mode.
	 * @param first         the index of the first element in the enabled arrays.
	 * @param count         the number of elements per instance.
	 * @param instanceCount the number of instances.
	 * @param baseInstance  the index of the first instance.
	 */
	public static void glDrawArraysInstancedBaseInstance(GLPrimitiveType mode, int first, int count, int instanceCount
			, int baseInstance) {
		GuiGlobals.GL.glDrawArraysInstancedBaseInstanceEXT(mode, first, count, instanceCount, baseInstance);
	}

	public static int glGetInteger(GLGetPName pname) {
		return GuiGlobals.GL.glGetInteger(pname);
	}

	public static void checkBoundBuffer(GLGetPName pname, int id) {
		if (GLFlags.EXPENSIVE_CHECKS) {
			int boundId = glGetInteger(pname);
			if (boundId != id) {
				throw new IllegalStateException(STR."Attempted to upload while not bound: expected \{id} found, \{boundId}");
			}
		}
	}

	/**
	 * @return The location of the uniform, or {@code -1} if it is not part of the specified program.
	 */
	public static int glGetUniformLocation(int program, String name) {
		return GuiGlobals.GL.glGetUniformLocation(program, name);
	}

	public static void glUniform1i(int location, int value) {
		GuiGlobals.GL.glUniform1i(location, value);
	}

	public static void glUniform4iv(int location, MemorySegment buffer) {
		GuiGlobals.GL.glUniform4iv(location, (int) (buffer.byteSize() / JAVA_INT.byteSize()), new ConstPointer<>(buffer));
	}

	public static void glEnable(GLEnableCap target) {
		GuiGlobals.GL.glEnable(target);
	}

	public static void glDisable(GLEnableCap target) {
		GuiGlobals.GL.glDisable(target);
	}

	public static void glClear(int mask) {
		GuiGlobals.GL.glClear(mask);
	}

	public static void glBindTexture(GLTextureTarget target, int id) {
		GuiGlobals.GL.glBindTexture(target, id);
	}

	public static int glGenTextures() {
		return GuiGlobals.GL.glGenTexture();
	}

	public static void glDeleteTextures(int handle) {
		GuiGlobals.GL.glDeleteTexture(handle);
	}

	/**
	 * @see <a href="https://docs.gl/gl4/glTexStorage3D">docs.gl</a>
	 * @since OpenGL 4.2
	 */
	public static void glTexStorage3D(GLTextureTarget target, int mipmapLevels, GLSizedInternalFormat format,
									  int width, int height, int depth) {
		GuiGlobals.GL.glTexStorage3D(target, mipmapLevels, format, width, height, depth);
	}

	/**
	 * @see <a href="https://docs.gl/gl4/glTexSubImage3D">docs.gl</a>
	 * @since OpenGL 2.0
	 */
	public static void glTexSubImage3D(GLTextureTarget target, int level, int xOffset, int yOffset, int zOffset,
									   int width, int height, int depth, GLPixelFormat format, GLPixelType type,
									   MemorySegment pixels) {
		if (GLFlags.CHECKS) {
			var length = (long) width * height * depth;
			if (pixels.byteSize() < length) {
				throw new IndexOutOfBoundsException(STR."Dimensions \{width},\{height},\{depth} (length of \{length}) overflow buffer of size \{pixels.byteSize()}");
			}
		}
		GuiGlobals.GL.glTexSubImage3D(target,
				level,
				xOffset,
				yOffset,
				zOffset,
				width,
				height,
				depth,
				format,
				type,
				new ConstPointer<>(pixels));
	}

	public static void glTexImage2D(GLTextureTarget target,
									int level,
									GLInternalFormat internalFormat,
									int width,
									int height,
									int border,
									GLPixelFormat format,
									GLPixelType type,
									MemorySegment pixels) {
		GuiGlobals.GL.glTexImage2D(target, level, internalFormat, width, height, border, format, type,
				new ConstPointer<>(pixels));
	}

	public static void glTexImage3D(GLTextureTarget target,
									int level,
									GLInternalFormat internalFormat,
									int width,
									int height,
									int depth,
									int border,
									GLPixelFormat format,
									GLPixelType type,
									MemorySegment pixels) {
		GuiGlobals.GL.glTexImage3D(target,
				level,
				internalFormat,
				width,
				height,
				depth,
				border,
				format,
				type,
				new ConstPointer<>(pixels));
	}

	public static void glTexParameteri(GLTextureTarget target, GLTextureParameterName param, int value) {
		GuiGlobals.GL.glTexParameteri(target, param, value);
	}
}
