package dev.pfaff.altargui.render;

import java.lang.foreign.MemorySegment;
import java.lang.foreign.ValueLayout;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

public final class MemorySegmentSink implements Sink {
	private static final ValueLayout.OfInt JAVA_INT_UNALIGNED_LITTLE_ENDIAN = ValueLayout.JAVA_INT_UNALIGNED.withOrder(ByteOrder.LITTLE_ENDIAN);
	private static final ValueLayout.OfInt JAVA_INT_UNALIGNED_BIG_ENDIAN = ValueLayout.JAVA_INT_UNALIGNED.withOrder(ByteOrder.BIG_ENDIAN);

	private final MemorySegment mem;
	private long offset;

	public MemorySegmentSink(MemorySegment mem, long offset) {
		this.mem = mem;
		this.offset = offset;
	}

	public MemorySegmentSink(MemorySegment mem) {
		this(mem, 0L);
	}

	private long remaining() {
		return mem.byteSize() - offset;
	}

	@Override
	public void checkRemaining(long length) {
		if (GLFlags.CHECKS && remaining() < length) {
			throw new BufferOverflowException(STR."Overflowed buffer by \{length - remaining()} bytes");
		}
	}

	@Override
	public void accept(byte value) {
		checkRemaining(ValueLayout.JAVA_BYTE.byteSize());
		mem.set(ValueLayout.JAVA_BYTE, offset, value);
		offset += ValueLayout.JAVA_BYTE.byteSize();
	}

	@Override
	public void accept(short value) {
		checkRemaining(ValueLayout.JAVA_SHORT_UNALIGNED.byteSize());
		mem.set(ValueLayout.JAVA_SHORT_UNALIGNED, offset, value);
		offset += ValueLayout.JAVA_SHORT_UNALIGNED.byteSize();
	}

	@Override
	public void accept(char value) {
		checkRemaining(ValueLayout.JAVA_CHAR_UNALIGNED.byteSize());
		mem.set(ValueLayout.JAVA_CHAR_UNALIGNED, offset, value);
		offset += ValueLayout.JAVA_CHAR_UNALIGNED.byteSize();
	}

	@Override
	public void accept(int value) {
		checkRemaining(ValueLayout.JAVA_INT_UNALIGNED.byteSize());
		mem.set(ValueLayout.JAVA_INT_UNALIGNED, offset, value);
		offset += ValueLayout.JAVA_INT_UNALIGNED.byteSize();
	}

	@Override
	public void accept(ByteOrder order, int value) {
		checkRemaining(ValueLayout.JAVA_INT_UNALIGNED.byteSize());
		mem.set(order == ByteOrder.LITTLE_ENDIAN ? JAVA_INT_UNALIGNED_LITTLE_ENDIAN : JAVA_INT_UNALIGNED_BIG_ENDIAN, offset, value);
		offset += ValueLayout.JAVA_INT_UNALIGNED.byteSize();
	}

	@Override
	public void accept(long value) {
		checkRemaining(ValueLayout.JAVA_LONG_UNALIGNED.byteSize());
		mem.set(ValueLayout.JAVA_LONG_UNALIGNED, offset, value);
		offset += ValueLayout.JAVA_LONG_UNALIGNED.byteSize();
	}

	@Override
	public void accept(float value) {
		checkRemaining(ValueLayout.JAVA_FLOAT_UNALIGNED.byteSize());
		mem.set(ValueLayout.JAVA_FLOAT_UNALIGNED, offset, value);
		offset += ValueLayout.JAVA_FLOAT_UNALIGNED.byteSize();
	}

	@Override
	public void accept(double value) {
		checkRemaining(ValueLayout.JAVA_DOUBLE_UNALIGNED.byteSize());
		mem.set(ValueLayout.JAVA_DOUBLE_UNALIGNED, offset, value);
		offset += ValueLayout.JAVA_DOUBLE_UNALIGNED.byteSize();
	}

	@Override
	public void accept(MemorySegment other) {
		checkRemaining(other.byteSize());
		mem.asSlice(offset).copyFrom(other);
		offset += other.byteSize();
	}

	@Override
	public void accept(ByteBuffer buffer, int offset, int length) {
		int restorePos = buffer.position();
		buffer.position(0);
		accept(MemorySegment.ofBuffer(buffer).asSlice(offset, length));
		buffer.position(restorePos);
	}

	public MemorySegment memorySegment() {
		return this.mem;
	}

	public long offset() {
		return this.offset;
	}

	public void advance(long offset) {
		this.offset += offset;
	}

	@Override
	public String toString() {
		return STR."MemorySegmentSink[mem=\{mem}]";
	}

}
