package dev.pfaff.altargui.render;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.render.objects.StaticGPUBuffer;
import dev.pfaff.altargui.uploadables.Uploadable;
import dev.pfaff.altargui.function.InfallibleAutoClosable;
import dev.pfaff.glass.opengl.enums.GLBufferTarget;

import java.io.Closeable;
import java.util.List;

import static dev.pfaff.altargui.GuiGlobals.LOGGER;
import static dev.pfaff.altargui.render.GLFlags.ALWAYS_UNBIND;
import static dev.pfaff.log4truth.StandardTags.ERROR;

public abstract class RawBuffer implements Closeable {
	protected final GLBufferTarget bindTarget;
	private int length;

	/**
	 * @param bindTarget      where the buffer will be bound.
	 */
	public RawBuffer(GLBufferTarget bindTarget) {
		this.bindTarget = bindTarget;
	}

	public static RawBuffer create(GLBufferTarget bindTarget, boolean isShared) {
		// TODO: allow changing this in a config file (and GUI)
		return new ViaIntermediateBuffer(bindTarget, isShared);
//		return new PersistentlyMappedBuffer(bindTarget, isShared);
//		return new MappedBuffer(bindTarget, isShared);
	}

	public final void bind() {
		GuiGlobals.GL.glBindBuffer(bindTarget, id());
	}

	public final void unbind() {
		if (ALWAYS_UNBIND) GuiGlobals.GL.glBindBuffer(bindTarget, 0);
	}

	/**
	 * Grows the buffer to at least the specified capacity. The buffer's contents are undefined after this operation.
	 */
	protected abstract void growBuffer(int capacity, Runnable onRecreate);

	protected abstract void resetNoGrow();

	/**
	 * @param capacity the required capacity in instances.
	 */
	public final void reset(int capacity, Runnable onRecreate) {
		if (capacity > this.capacity()) {
			growBuffer(capacity, onRecreate);
		} else {
			resetNoGrow();
		}

		length = 0;
	}

	/**
	 * Begins uploading to the buffer. Attempts to upload more than {@code capacity} might result in a
	 * {@link BufferOverflowException} being thrown.
	 * <p>
	 * The buffer's contents are undefined after this operation.
	 */
	public void beginUploading() {
	}

	public void finishUploading() {
	}

	public abstract boolean isBoundDuringUpload();

	public final InfallibleAutoClosable boundIfNotAlreadyDuringUpload() {
		if (!isBoundDuringUpload()) {
			bind();
			return this::unbind;
		} else {
			return () -> {};
		}
	}

	/**
	 * @param addit the number of additional bytes.
	 */
	private void checkUpload(long offset, int addit) {
		// check the remaining capacity
		if (GLFlags.CHECKS && offset + addit > capacity()) {
			throw new BufferOverflowException(STR."Offset \{offset} plus additional bytes \{addit} overflows capacity \{capacity()}");
		}
		// check that the buffer is still bound
		//GL.checkBoundBuffer(bindTargetPname, id);
	}

//	private final void uploadAt(long offset, StaticGPUBuffer buffer) {
//		int addit = buffer.stride() * buffer.count();
//		checkUpload(offset, addit);
//
////		LOGGER.log(Level.INFO, "uploading " + addit + " bytes");
//
//		GL.glBindBuffer(GL31.GL_COPY_READ_BUFFER, buffer.id());
//
//		GL.glCopyBufferSubData(GL31.GL_COPY_READ_BUFFER,
//							   bindTarget,
//							   0,
//							   length,
//							   (long) buffer.stride() * buffer.count());
//
//		if (ALWAYS_UNBIND) GL.glBindBuffer(GL31.GL_COPY_READ_BUFFER, 0);
//
//		this.length += addit;
//	}

	protected final void checkCompleteUpload(long uploaded, Uploadable uploadable) {
		if (GLFlags.CHECKS) {
			int expected = uploadable.size();
			if (uploaded != expected) {
				throw new BufferOverflowException(STR."Expected the uploadable \{uploadable} to upload \{expected} bytes, but \{uploaded < expected
						? "only"
						: "instead"} \{uploaded} bytes were uploaded.");
			}
		}
	}

	protected abstract void uploadAtUnchecked(long offset, Uploadable uploadable);

	public final void uploadAt(long offset, Uploadable uploadable) {
		if (uploadable instanceof StaticGPUBuffer) {
			// TODO: special case in uploadAtUnchecked in sub-classes
		}
		checkUpload(offset, uploadable.size());
		uploadAtUnchecked(offset, uploadable);
	}

	public final void upload(Uploadable uploadable) {
		uploadAt(this.length, uploadable);
		this.length += uploadable.size();
	}

	public abstract int id();

	public final int length() {
		return length;
	}

	/**
	 * Unsafe.
	 */
	public final void setLength(int length) {
		this.length = length;
	}

	public abstract int capacity();

	@Override
	public String toString() {
		return STR."\{getClass().getSimpleName()}[id=\{id()}]";
	}

	public void dump(String label, boolean capacity) {
		LOGGER.log(() -> STR."\{label}:  dump not implemented", List.of(ERROR));
	}

	@Override
	public void close() {
		GLUtil.glDeleteBuffers(id());
	}
}
