package dev.pfaff.altargui.resource;

import org.jspecify.annotations.NonNull;

import java.io.IOException;
import java.io.InputStream;
import java.io.UncheckedIOException;
import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.SegmentAllocator;
import java.net.URI;
import java.net.http.HttpRequest;
import java.net.http.HttpResponse;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.CancellationException;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.CompletionStage;
import java.util.concurrent.Flow;

public record UrlLoadableResource(UrlResourceLoader loader, String url) implements LoadableResource {
	@Override
	public @NonNull MemorySegment readBytes() throws IOException {
		return doReqRes(responseInfo -> {
			checkResponseInfo(responseInfo);
			return new MemorySegmentSubscriber(Arena.ofAuto());
		});
	}

	@Override
	public @NonNull InputStream openInputStream() throws IOException {
		return doReqRes(responseInfo -> {
			checkResponseInfo(responseInfo);
			return HttpResponse.BodySubscribers.ofInputStream();
		});
	}

	/**
	 * @throws IOException if an I/O error occurs
	 * @throws CancellationException if the operation is interrupted
	 */
	private <T> T doReqRes(HttpResponse.BodyHandler<T> bodyHandler) throws IOException, CancellationException {
		var req = HttpRequest.newBuilder().GET().uri(URI.create(this.url)).build();
		try {
			return this.loader.client.send(req, bodyHandler).body();
		} catch (InterruptedException e) {
			throw new CancellationException(e.getMessage());
		}
	}

	private static void checkResponseInfo(HttpResponse.ResponseInfo responseInfo) {
		if (responseInfo.statusCode() != 200) {
			throw new UncheckedIOException(new IOException(STR."Non-OK status code \{responseInfo.statusCode()}"));
		}
	}

	private static final class MemorySegmentSubscriber implements HttpResponse.BodySubscriber<MemorySegment> {
		private final SegmentAllocator allocator;

		private final CompletableFuture<MemorySegment> result = new CompletableFuture<>();
		private final List<ByteBuffer> received = new ArrayList<>();

		private volatile Flow.Subscription subscription;

		public MemorySegmentSubscriber(SegmentAllocator allocator) {
			this.allocator = Objects.requireNonNull(allocator);
		}

		@Override
		public void onSubscribe(Flow.Subscription subscription) {
			if (this.subscription != null) {
				subscription.cancel();
				return;
			}
			this.subscription = subscription;
			// We can handle whatever you've got
			subscription.request(Long.MAX_VALUE);
		}

		@Override
		public void onNext(List<ByteBuffer> items) {
			// incoming buffers are allocated by http client internally,
			// and won't be used anywhere except this place.
			// So it's free simply to store them for further processing.
			assert hasRemaining(items);
			received.addAll(items);
		}

		@Override
		public void onError(Throwable throwable) {
			received.clear();
			result.completeExceptionally(throwable);
		}

		private static MemorySegment join(SegmentAllocator allocator, List<ByteBuffer> bytes) {
			int size = remaining(bytes, Integer.MAX_VALUE);
			MemorySegment res = allocator.allocate(size);
			int from = 0;
			for (ByteBuffer b : bytes) {
				int l = b.remaining();
				var src = MemorySegment.ofBuffer(b);
				res.asSlice(from).copyFrom(src);
				from += l;
			}
			return res;
		}

		@Override
		public void onComplete() {
			try {
				result.complete(join(allocator, received));
				received.clear();
			} catch (IllegalArgumentException e) {
				result.completeExceptionally(e);
			}
		}

		@Override
		public CompletionStage<MemorySegment> getBody() {
			return result;
		}

		private static boolean hasRemaining(List<ByteBuffer> bufs) {
			for (ByteBuffer buf : bufs) {
				if (buf.hasRemaining())
					return true;
			}
			return false;
		}

		private static int remaining(List<ByteBuffer> bufs, int max) {
			long remain = 0;
			for (ByteBuffer buf : bufs) {
				remain += buf.remaining();
				if (remain > max) {
					throw new IllegalArgumentException("too many bytes");
				}
			}
			return (int) remain;
		}
	}
}
