package dev.pfaff.altargui.texture;

import dev.pfaff.altargui.function.InfallibleAutoClosable;
import dev.pfaff.altargui.io.IOUtil;
import dev.pfaff.altargui.libimage.*;
import dev.pfaff.altargui.render.GLUtil;
import dev.pfaff.altargui.util.BasicFuture;
import dev.pfaff.altargui.util.GLThread;
import dev.pfaff.altargui.util.ThreadFuture;
import dev.pfaff.glass.opengl.MutPointer;
import dev.pfaff.glass.opengl.enums.*;
import dev.pfaff.log4truth.Logger;
import dev.pfaff.log4truth.NamedLogger;
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.io.IOException;
import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.ref.SoftReference;
import java.nio.file.Path;
import java.util.BitSet;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.Callable;
import java.util.concurrent.Semaphore;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.render.GLContextPool.GL_CONTEXT_POOL;
import static dev.pfaff.altargui.render.GLContextPool.POOL_ACQUIRED_GL;
import static dev.pfaff.altargui.util.BasicFuture.completedTask;
import static dev.pfaff.altargui.util.BasicFuture.executeOn;
import static dev.pfaff.altargui.util.ThreadFuture.spawnVirtualTask;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static dev.pfaff.log4truth.StandardTags.ERROR;
import static java.lang.foreign.ValueLayout.JAVA_BYTE;
import static java.lang.foreign.ValueLayout.JAVA_INT;

public final class TextureAllocator implements AutoCloseable {
	private static final NamedLogger LOGGER = new NamedLogger("TexAlloc");

	public static final int MAX_TEXTURE_SIZE;

	static {
		// cannot use lambdas across threads in static initializer
		MAX_TEXTURE_SIZE = executeOn(new Callable<Integer>() {
			@Override
			public Integer call() throws Exception {
				return GL.glGetInteger(GLGetPName.MAX_TEXTURE_SIZE);
			}
		}, GLThread::submitToGLThread);
//		MAX_TEXTURE_SIZE = GL.glGetInteger(GLGetPName.MAX_TEXTURE_SIZE);
		if (MAX_TEXTURE_SIZE == 0) {
			throw new IllegalStateException("MAX_TEXTURE_SIZE must be greater than 0");
		}
	}

	/**
	 * The minimum allocation size in pixels. Herein we often measure in "units" rather than pixels for efficiency.
	 */
//	public static final int UNIT = 8;
	public static final int UNIT = Math.divideExact(MAX_TEXTURE_SIZE, 256);

	static {
		if (Integer.bitCount(UNIT) != 1) {
			throw new UnsupportedOperationException(STR."Only power-of-two unit sizes are supported, but \{UNIT} was calculated");
		}
	}

	/**
	 * The size (in {@link #UNIT}s) of each layer in the allocator.
	 */
	private final int sizeUnits;
	private int layers;
	private boolean ab;
	private final TextureArray textureArrayA = new TextureArray();
	private final TextureArray textureArrayB = new TextureArray();
	private final BitSet allocations = new BitSet();
	private final Object2ObjectOpenHashMap<LoadableTexture, Object> cache =
			new Object2ObjectOpenHashMap<>();
	private final Semaphore lock = new Semaphore(1);

	/**
	 * @param size the width and height of each layer.
	 */
	public TextureAllocator(int size, int initialLayers) {
		if (size > MAX_TEXTURE_SIZE) {
			throw new IllegalArgumentException("Layer size too large");
		}
		this.sizeUnits = size / UNIT;
		if (size % UNIT != 0) throw new IllegalArgumentException(STR."Invalid layer size: \{size}");
		this.layers = initialLayers;
	}

	public void reset() {
		this.reset(this.getArray());
	}

	private void reset(TextureArray textureArray) {
		var size = this.sizeUnits * UNIT;
		textureArray.allocate(size, size, this.layers, TextureOptions.DEFAULT);
	}

	private void flip() {
		this.ab = !this.ab;
	}

	private TextureArray getArray() {
		return this.ab ? this.textureArrayB : this.textureArrayA;
	}

	private TextureArray getBackArray() {
		return this.ab ? this.textureArrayB : this.textureArrayA;
	}

	public InfallibleAutoClosable bound() {
		return this.getArray().bound();
	}

	/**
	 * <p>
	 * Gets a shader to the given texture. If not loaded, the texture will be loaded and a slot allocated for it.
	 * <p>
	 * This method must only be called from the main thread.
	 */
	public BasicFuture<TextureHandle> alloc(LoadableTexture key) {
		{
			var cacheValue = this.cache.get(key);
			// this inspection is wrong
			//noinspection ConstantValue
			if (cacheValue != null) {
				var handle = tryDerefCacheValue(cacheValue);
				if (handle != null) return handle;
			}
		}
		BasicFuture<TextureHandle> future = spawnVirtualTask(() -> {
			ImageData image;
			try {
				image = key.load();
			} catch (Throwable t) {
				LOGGER.log(() -> "Texture load failed", t, List.of(ERROR));
				return Objects.requireNonNull(tryDerefCacheValueJoin(this.cache.get(SinglePixelLoadableTexture.WHITE)));
			}

			try {
				// allocate a new image in the texture array.
				var handle = this.allocNew(key, image);

				// try getting from the cache again before storing to avoid keeping alive duplicate allocations.
				executeOn(() -> {
					assert this.cache.get(key) instanceof ThreadFuture<?>;
					this.cache.put(key, new SoftReference<>(handle));
				}, GLThread::submitToGLThread);
				return handle;
			} catch (Throwable t) {
				LOGGER.log(() -> "Texture allocation failed", t, List.of(ERROR));
				return Objects.requireNonNull(tryDerefCacheValueJoin(this.cache.get(SinglePixelLoadableTexture.WHITE)));
			}
		});
		this.cache.put(key, future);
		return future;
	}

	private static @Nullable BasicFuture<TextureHandle> tryDerefCacheValue(Object cacheValue) {
		if (cacheValue instanceof SoftReference<?> ref) {
			var handle = (TextureHandle) ref.get();
			if (handle != null) return completedTask(handle);
		} else {
			//noinspection unchecked
			return (BasicFuture<TextureHandle>) cacheValue;
		}
		return null;
	}

	private static @Nullable TextureHandle tryDerefCacheValueJoin(Object cacheValue) {
		if (cacheValue instanceof SoftReference<?> ref) {
			return (TextureHandle) ref.get();
		} else {
			//noinspection unchecked
			return ((BasicFuture<TextureHandle>) cacheValue).join();
		}
	}

	private TextureHandle allocNew(LoadableTexture key, ImageData image) throws InterruptedException {
		int width = image.info().width();
		int height = image.info().height();

		// align size up to the next multiple of UNIT.
		// these are in units, not pixels
		int unitWidth = (width + (UNIT - 1)) / UNIT;
		int unitHeight = (height + (UNIT - 1)) / UNIT;
		assert unitWidth * UNIT >= width && unitHeight * UNIT >= height : STR."width and/or height was improperly aligned to unit size:\n\t\{width} > \{unitWidth * UNIT}\n\tor\n\t\{height} > \{unitHeight * UNIT}";

		int sizeUnits = this.sizeUnits;

		if (unitWidth > sizeUnits || unitHeight > sizeUnits) {
			throw new IllegalArgumentException(STR."Size of image exceeds size of allocator layers: \{width} || \{height} > \{this.sizeUnits * UNIT}");
		}

		LOGGER.log(() -> STR."Searching for a free region to allocate \{key} an image of \{width}x\{height}", List.of(DEBUG));

		DetailedHandle detailed;

		this.lock.acquire();
		try {
			// find a free region
			int allocationsSize = sizeUnits * sizeUnits * this.layers;
			int fromSearchUpr = allocationsSize - unitWidth - (unitHeight - 1) * sizeUnits;
			int layers = this.layers;
			int sizeUnitsSq = sizeUnits * sizeUnits;

			int xUpr = sizeUnits - unitHeight;

			int from = -1, iUpr = -1;

			search:
			for (int layerStart = 0; layerStart < layers * sizeUnitsSq; layerStart += sizeUnitsSq) {
				from = layerStart;
				int layerUpr = from + sizeUnitsSq;
				while (from <= fromSearchUpr) {
					assert from % sizeUnits == 0 :
							STR."not at the start of a row: from=\{from}";
					int rowUpr = from + xUpr;
					for (; from < rowUpr; from++) {
						iUpr = from + unitHeight * sizeUnits;
						assert iUpr <= layerUpr;

						if (isRegionFree(from, iUpr, unitWidth)) break search;
					}
					from += unitHeight;
				}
				iUpr = -1;
			}

			if (iUpr == -1) {
				throw new UnsupportedOperationException("TODO: realloc");
			}
			assert from != -1;

			detailed = this.makeHandle(from, iUpr, unitWidth, width, height);
		} finally {
			this.lock.release();
		}

		upload(image, detailed, width, height);
		return detailed.handle();
	}

	private void upload(ImageData image, DetailedHandle detailed, int width, int height) {
		var handle = detailed.handle();
		LOGGER.log(() -> STR."Allocated texture handle \{handle.layer}:\{handle.u},\{handle.v} for image of \{width}x\{height} at \{
				detailed.u() * UNIT},\{detailed.v() * UNIT}", List.of(DEBUG));
		var array = this.getArray();

		executeOn(() -> {
			var gl = POOL_ACQUIRED_GL.get();
			assert gl == GL;

			// upload
			try (var _ = this.bound()) {
				array.startUploading();
				array.upload(handle.layer,
						detailed.u() * UNIT,
						detailed.v() * UNIT,
						width,
						height,
						image.info().colorType().glPixelFormat(),
						image.info().colorType().glChannelType(),
						image.pixels());
			}
		}, GL_CONTEXT_POOL.acquireExecutor());
		LOGGER.log(() -> STR."Uploaded image to \{handle.layer}:\{handle.u},\{handle.v}", List.of(DEBUG));
	}

	private static @NonNull MappedBuffer genAndMapBuffer(ImageData image, dev.pfaff.glass.opengl.GL gl) {
		int id_ = gl.glGenBuffer();
		gl.glBindBuffer(GLBufferTarget.PIXEL_UNPACK_BUFFER, id_);
		gl.glBufferData(GLBufferTarget.PIXEL_UNPACK_BUFFER, image.pixels().byteSize(), null, GLBufferUsage.STREAM_DRAW);
		var segment_ = GLUtil.glMapBufferRange(GLBufferTarget.PIXEL_UNPACK_BUFFER,
				0L,
				image.pixels().byteSize(),
				GLMapBufferAccessMask.MAP_WRITE_BIT.glEnum()
				| GLMapBufferAccessMask.MAP_INVALIDATE_BUFFER_BIT.glEnum());
		gl.glBindBuffer(GLBufferTarget.PIXEL_UNPACK_BUFFER, 0);
		return new MappedBuffer(id_, segment_);
	}

	private DetailedHandle makeHandle(int from, int iUpr, int unitWidth, int textureWidth, int textureHeight) {
		var allocations = this.allocations;
		int sizeUnits = this.sizeUnits;

		// mark allocated
		for (int i = from; i < iUpr; i += sizeUnits) { // for each row
			// mark the region of each row allocated
			allocations.set(i, i + unitWidth);
		}

		var handle = MemorySegment.NULL.reinterpret(Arena.ofAuto(), _ -> {
			this.lock.acquireUninterruptibly();
			try {
				// mark free
				for (int i = from; i < iUpr; i += sizeUnits) {
					allocations.clear(i, i + unitWidth);
				}
			} finally {
				this.lock.release();
			}
		});

		int layer = from / (sizeUnits * sizeUnits);
		int u = from % sizeUnits;
		int v = (from / sizeUnits) % sizeUnits;
		return new DetailedHandle(handle, layer, sizeUnits, u, v, textureWidth, textureHeight);
	}

	private boolean isRegionFree(int from, int iUpr, int unitWidth) {
		var allocations = this.allocations;
		int sizeUnits = this.sizeUnits;

		// check if all are free
		for (int i = from; i < iUpr; i += sizeUnits) {
			for (int x = 0; x < unitWidth; x++) {
				if (allocations.get(i + x)) return false;
			}
		}
		return true;
	}

	@Override
	public void close() {
		this.textureArrayA.close();
		this.textureArrayB.close();
	}

	public void dump() {
		var arr = this.getArray();
		try {
			try (var _ = arr.bound()) {
				try (var arena = Arena.ofConfined()) {
					var segment = arena.allocate(JAVA_INT,
							(this.sizeUnits * UNIT) * (this.sizeUnits * UNIT * this.layers));
					GL.glGetTexImage(arr.target(),
							0,
							GLPixelFormat.RGBA,
							GLPixelType.UNSIGNED_BYTE,
							new MutPointer<>(segment));
					var encoded = LibImage.encodeImage(
							new ImageInfo(this.sizeUnits * UNIT, this.sizeUnits * UNIT * this.layers, ColorType.Rgba8),
							segment,
							ImageFormat.Png,
							EncoderProfile.Fast,
							arena);
					IOUtil.write(Path.of("allocated-texture.png"), encoded);
				}
			}

			try (var arena = Arena.ofConfined()) {
				var size = this.sizeUnits * this.sizeUnits * this.layers;
				var segment = arena.allocate(size);
				for (int i = 0; i < size; i++) {
					var pixel = this.allocations.get(i) ? 0xff : 0x00;
					segment.setAtIndex(JAVA_BYTE, i, (byte) pixel);
				}
				var encoded = LibImage.encodeImage(
						new ImageInfo(this.sizeUnits, this.sizeUnits * this.layers, ColorType.L8),
						segment,
						ImageFormat.Png,
						EncoderProfile.Fast,
						arena);
				IOUtil.write(Path.of("allocated-units.png"), encoded);
			}
		} catch (IOException e) {
			Logger.log(() -> "Failed to dump texture atlas", e, List.of(ERROR));
		}
	}

	/**
	 * @param u measured in {@link #UNIT}s
	 * @param v measured in {@link #UNIT}s
	 */
	record DetailedHandle(TextureHandle handle, int u, int v) {
		/**
		 * @param layer index
		 * @param layerSizeInUnits the size of each layer, measured in {@link #UNIT}s
		 * @param u measured in {@link #UNIT}s
		 * @param v measured in {@link #UNIT}s
		 */
		public DetailedHandle(MemorySegment handle, int layer, int layerSizeInUnits, int u, int v, int textureWidth, int textureHeight) {
			this(new TextureHandle(handle, layer, layerSizeInUnits * UNIT, u / (float)layerSizeInUnits, v / (float)layerSizeInUnits, textureWidth, textureHeight), u, v);
		}
	}

	private record MappedBuffer(int id, MemorySegment segment) {}
}
