package dev.pfaff.altargui.texture;

import dev.pfaff.altargui.function.InfallibleAutoClosable;
import dev.pfaff.altargui.render.GLUtil;
import dev.pfaff.glass.opengl.ConstPointer;
import dev.pfaff.glass.opengl.enums.*;

import java.lang.foreign.MemorySegment;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.render.GLFlags.ALWAYS_UNBIND;

public final class TextureArray implements AutoCloseable {
	public final int handle;
	private int width;
	private int height;
	private int layerCount;

	public TextureArray() {
		this.handle = GLUtil.glGenTextures();
	}

	public InfallibleAutoClosable bound() {
		GL.glActiveTexture(GLTextureUnit.TEXTURE0);
		GLUtil.glBindTexture(target(), handle);
		return ALWAYS_UNBIND ? () -> GLUtil.glBindTexture(target(), 0) : () -> {};
	}

	/**
	 * Allocates the texture. It must already be {@link #bound}.
	 */
	public void allocate(int width, int height, int layerCount, TextureOptions options) {
		this.width = width;
		this.height = height;
		this.layerCount = layerCount;

		if (layerCount != 1) throw new AssertionError();
		GL.glTexStorage2D(target(), 1, GLSizedInternalFormat.RGBA8, width, height);

		GLUtil.glTexParameteri(target(), GLTextureParameterName.TEXTURE_MIN_FILTER, options.minFilter().glEnum());
		GLUtil.glTexParameteri(target(), GLTextureParameterName.TEXTURE_MAG_FILTER, options.magFilter().glEnum());
		if (options.clamp()) {
			GLUtil.glTexParameteri(target(), GLTextureParameterName.TEXTURE_WRAP_S, GLTextureWrapMode.CLAMP_TO_EDGE.glEnum());
			GLUtil.glTexParameteri(target(), GLTextureParameterName.TEXTURE_WRAP_T, GLTextureWrapMode.CLAMP_TO_EDGE.glEnum());
		}
	}

	public void startUploading() {
		GL.glPixelStorei(GLPixelStoreParameter.UNPACK_ROW_LENGTH, 0);
		GL.glPixelStorei(GLPixelStoreParameter.UNPACK_SKIP_PIXELS, 0);
		GL.glPixelStorei(GLPixelStoreParameter.UNPACK_SKIP_ROWS, 0);
		GL.glPixelStorei(GLPixelStoreParameter.UNPACK_ALIGNMENT, 4);
	}

	/**
	 * Remember to call {@link #startUploading()}.
	 */
	public void upload(int layer, int xOffset, int yOffset, int width, int height, GLPixelFormat format, GLPixelType channelType, MemorySegment pixels) {
		if (0 > xOffset || 0 > yOffset || xOffset > this.width || yOffset > this.height) {
			throw new IllegalArgumentException(STR."Upload offset out of bounds: \{xOffset} !in [0, \{this.width}] || \{yOffset} !in [0, \{this.height}]");
		}
		var availWidth = this.width - xOffset;
		var availHeight = this.height - yOffset;
		if (0 > width || 0 > height || width > availWidth || height > availHeight) {
			throw new IllegalArgumentException(STR."Upload size out of bounds: \{width} !in [0, \{availWidth}] || \{height} !in [0, \{availHeight}]");
		}
		if (layer >= this.layerCount) {
			throw new IllegalArgumentException(STR."Upload layer out of bounds: \{layer} >= \{this.layerCount}");
		}

		if (layer != 0) throw new AssertionError();
		GL.glTexSubImage2D(target(),
				0,
				xOffset,
				yOffset,
				width,
				height,
				format,
				channelType,
				new ConstPointer<>(pixels));
	}

	/**
	 * Remember to call {@link #startUploading()}.
	 */
	public void uploadFromPixelUnpackBuffer(int layer,
											int xOffset, int yOffset,
											int width, int height,
											GLPixelFormat format,
											GLPixelType channelType) {
		if (0 > xOffset || 0 > yOffset || xOffset > this.width || yOffset > this.height) {
			throw new IllegalArgumentException(STR."Upload offset out of bounds: \{xOffset} !in [0, \{this.width}] || \{yOffset} !in [0, \{this.height}]");
		}
		var availWidth = this.width - xOffset;
		var availHeight = this.height - yOffset;
		if (0 > width || 0 > height || width > availWidth || height > availHeight) {
			throw new IllegalArgumentException(STR."Upload size out of bounds: \{width} !in [0, \{availWidth}] || \{height} !in [0, \{availHeight}]");
		}
		if (layer >= this.layerCount) {
			throw new IllegalArgumentException(STR."Upload layer out of bounds: \{layer} >= \{this.layerCount}");
		}

		if (layer != 0) throw new AssertionError();
		GL.glTexSubImage2D(target(),
				0,
				xOffset,
				yOffset,
				width,
				height,
				format,
				channelType,
				null);
	}

	/**
	 * Remember to call {@link #startUploading()}.
	 */
	public void upload(int layer, int width, int height, GLPixelFormat format, GLPixelType channelType, MemorySegment pixels) {
		this.upload(layer, 0, 0, width, height, format, channelType, pixels);
	}

	/**
	 * Remember to call {@link #startUploading()}.
	 */
	public void uploadFull(GLPixelFormat format, MemorySegment pixels) {
		long expected = (long) width * height * layerCount * 4;
		if (pixels.byteSize() != expected) {
			throw new IllegalArgumentException(STR."Pixel buffer size is wrong: \{pixels.byteSize()} != \{expected}");
		}

		GL.glTexSubImage2D(target(),
				0,
				0,
				0,
				width,
				height,
				format,
				GLPixelType.UNSIGNED_BYTE,
				new ConstPointer<>(pixels));
	}

	@Override
	public void close() {
		GLUtil.glDeleteTextures(this.handle);
	}

	public GLTextureTarget target() {
		return GLTextureTarget.TEXTURE_2D;
	}
}
