package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.internal.util.StencilLayoutContext;
import dev.pfaff.altargui.types.Alignment;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;

/**
 * @param aspectRatio the desired aspect ratio, in the form {@code width / height}.
 * @param horizontalAlignment the horizontal alignment if the width is increased to match the height.
 * @param verticalAlignment the vertical alignment if the height is increased to match the width.
 */
public record AspectRatioWidget(Widget child, double aspectRatio, Alignment horizontalAlignment, Alignment verticalAlignment) implements SingleChildWidget {
	public AspectRatioWidget {
		Objects.requireNonNull(child, "child cannot be null");
		Objects.requireNonNull(horizontalAlignment, "horizontalAlignment cannot be null");
		Objects.requireNonNull(verticalAlignment, "verticalAlignment cannot be null");
	}

	@Override
	public @NotNull AbstractState<AspectRatioWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends AbstractState<AspectRatioWidget> {
		private double aspectRatio;
		private Alignment horizontalAlignment = Alignment.Start;
		private Alignment verticalAlignment = Alignment.Start;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, AspectRatioWidget widget, boolean init) {
			super.configure(stateContext, widget, init);
			if (this.aspectRatio != widget.aspectRatio) {
				this.aspectRatio = widget.aspectRatio;
				this.markNeedsStencilLayout();
			}
			boolean traceDirty = false;
			if (this.horizontalAlignment != widget.horizontalAlignment) {
				this.horizontalAlignment = widget.horizontalAlignment;
				traceDirty = true;
			}
			if (this.verticalAlignment != widget.verticalAlignment) {
				this.verticalAlignment = widget.verticalAlignment;
				traceDirty = true;
			}
			if (traceDirty) this.markNeedsTraceLayout();
		}

		@Override
		protected void computeStencilLayout(StateContext context) {
			var allocWidth = allocatedWidth();
			var allocHeight = allocatedHeight();
			var aspectRatio = this.aspectRatio;

			float width, height;
			{
				var aspectRatioWidthFromHeight = Math.min((float) (allocHeight * aspectRatio), allocWidth);
				// shrink to match
				if (aspectRatioWidthFromHeight > allocWidth) {
					var aspectRatioHeightFromWidth = (float) (allocWidth / aspectRatio);
					assert aspectRatioHeightFromWidth <= allocHeight :
							STR."the math is not mathing: allocatedWidth=\{allocWidth}, allocatedHeight=\{allocHeight}, aspectRatio=\{aspectRatio}";
					width = allocWidth;
					height = aspectRatioHeightFromWidth;
				} else {
					width = aspectRatioWidthFromHeight;
					height = allocHeight;
				}
			}

			var childElement = getChildElement();
			StencilLayoutContext.updateMountAndLayout(context, this, childElement, (e, ctx) -> {
				e.setAvailableSize(width, height);
				e.stencilLayout(ctx);
			});

			float cwidth, cheight;
			{
				var compWidth = childElement.computedWidth();
				var compHeight = childElement.computedHeight();
				var aspectRatioWidthFromHeight = (float) (compHeight * aspectRatio);
				// expand to match
				if (aspectRatioWidthFromHeight <= compWidth) {
					var aspectRatioHeightFromWidth = (float) (compWidth / aspectRatio);
					assert aspectRatioHeightFromWidth >= compHeight :
							STR."the math is not mathing: computedWidth=\{compWidth}, computedHeight=\{compHeight}, aspectRatio=\{aspectRatio}";
					cwidth = compWidth;
					cheight = aspectRatioHeightFromWidth;
				} else {
					cwidth = aspectRatioWidthFromHeight;
					cheight = compHeight;
				}
			}
			setComputedSize(cwidth, cheight);
		}

		@Override
		protected void computeTraceLayout() {
			var e = getChildElement();
			e.setPosition(allocatedX() + this.horizontalAlignment.apply(e.computedWidth(), this.computedWidth()),
					allocatedY() + this.verticalAlignment.apply(e.computedHeight(), this.computedHeight()));
			e.traceLayout();
		}
	}
}
