package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.internal.util.StencilLayoutContext;
import org.jetbrains.annotations.NotNull;

/**
 * A widget that calls the provided {@code transformer} every layout.
 */
public record CustomTransformBuilder(Widget child, Transformer transformer) implements SingleChildWidget {
	@Override
	public @NotNull AbstractState<CustomTransformBuilder> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	public static final class Transform {
		public float x;
		public float y;

		private Transform(float x, float y) {
			this.x = x;
			this.y = y;
		}
	}

	@FunctionalInterface
	public interface Transformer {
		void transform(Transform transform,
					   float computedWidth, float computedHeight,
					   float availableWidth, float availableHeight);
	}

	private static final class State extends AbstractState<CustomTransformBuilder> {
		private Transformer transformer;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, CustomTransformBuilder widget, boolean init) {
			super.configure(stateContext, widget, init);
			if (!widget.transformer.equals(this.transformer)) {
				this.transformer = widget.transformer;
				markNeedsTraceLayout();
			}
		}

		@Override
		protected void computeStencilLayout(StateContext context) {
			// TODO: is it ideal to copy the child's size?
			super.computeStencilLayout(context);
		}

		@Override
		protected void computeTraceLayout() {
			var e = getChildElement();
			var transform = new Transform(this.allocatedX(), this.allocatedY());
			this.transformer.transform(transform,
					e.computedWidth(),
					e.computedHeight(),
					e.allocatedWidth(),
					e.allocatedHeight());
			if (e.allocatedX() != transform.x || e.allocatedY() != transform.y) {
				e.setPosition(transform.x, transform.y);
				markNeedsScissorLayout();
			}

			e.traceLayout();
		}

		@Override
		protected void computeScissorLayout() {
			var childElement = this.getChildElement();
			float allocatedX = childElement.allocatedX();
			float allocatedY = childElement.allocatedY();
			var x = Math.max(allocatedX, 0);
			var y = Math.max(allocatedY, 0);
			childElement.setClippedRect(x,
					y,
					Math.max(childElement.computedWidth() + (allocatedX - x), 0),
					Math.max(childElement.computedHeight() + (allocatedY - y), 0));
			childElement.scissorLayout();
		}
	}
}
