package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.function.IndexedFunction;
import dev.pfaff.altargui.handlers.MouseScrollHandler;
import dev.pfaff.altargui.inspector.DebugLayer;
import org.jetbrains.annotations.NotNull;

import java.util.List;

import static dev.pfaff.altargui.Key.intKey;
import static dev.pfaff.altargui.util.ClipUtil.clipExtent;

public record FiniteScrollListWidget(int count, Builder builder) implements Widget {
	public static final DebugLayer DEBUG_LAYER_PRE_UNSCISSORED = new DebugLayer(FiniteScrollListWidget.class, "pre_unscissored");
	public static final DebugLayer DEBUG_LAYER_PRE_SCISSORED = new DebugLayer(FiniteScrollListWidget.class, "pre_scissored");
	public static final DebugLayer DEBUG_LAYER_POST_UNSCISSORED = new DebugLayer(FiniteScrollListWidget.class, "post_unscissored");

	public FiniteScrollListWidget(int count, KeyFunction keyFunction, WidgetFunction builder) {
		this(count, (context, i) -> context.layout(keyFunction.apply(i), builder.build(i)));
	}

	public FiniteScrollListWidget(int count, WidgetFunction builder) {
		this(count, (context, i) -> context.layout(intKey(i), builder.build(i)));
	}

	public static <T> FiniteScrollListWidget create(List<T> items,
													IndexedFunction<T, Key.CallSite> keyFunction,
													IndexedFunction<T, Widget> builder) {
		return new FiniteScrollListWidget(items.size(), (context, i) -> {
			T item = items.get(i);
			context.layout(keyFunction.apply(i, item), builder.apply(i, item));
		});
	}

	public static <T> FiniteScrollListWidget create(List<T> items,
													IndexedFunction<T, Widget> builder) {
		return new FiniteScrollListWidget(items.size(), (context, i) -> {
			context.layout(intKey(i), builder.apply(i, items.get(i)));
		});
	}

	@Override
	public @NotNull WidgetInstance<FiniteScrollListWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	@FunctionalInterface
	public interface KeyFunction {
		Key.CallSite apply(int index);
	}

	@FunctionalInterface
	public interface WidgetFunction {
		Widget build(int index);
	}

	@FunctionalInterface
	public interface Builder {
		/**
		 * @param context the context to accept the built widget.
		 * @param i the index of the widget to build.
		 */
		void build(LayoutContext context, int i);
	}

	private static final class State extends WidgetInstance<FiniteScrollListWidget> implements MouseScrollHandler {
		private int count;
		private Builder builder;

		/**
		 * The index of the first item displayed.
		 */
		private int itemI;

		/**
		 * The offset of the first item displayed. Negative values shift the items above the viewport, while positive
		 * values trigger the {@link #itemI} to be decremented and the offset updated accordingly until it is at most
		 * {@code 0.0f}.
		 */
		private double itemOffset;

		private final WidgetInstanceList children = new WidgetInstanceList(2);

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, FiniteScrollListWidget widget, boolean init) {
			var dirtyStencil = false;
			if (this.count != widget.count) {
				this.count = widget.count;
				dirtyStencil = true;
			}
			if (!widget.builder.equals(this.builder)) {
				this.builder = widget.builder;
				dirtyStencil = true;
			}

			if (dirtyStencil) markNeedsStencilLayout();
		}

		// TODO: cache built widgets
		@Override
		protected void computeStencilLayout(StateContext context) {
			markNeedsTraceLayout();

			var children = this.children;
			children.clear();

			var builder = this.builder;
			int count = this.count;
			final var allocWidth = this.allocatedWidth();
			final var allocHeight = this.allocatedHeight();
			float width = 0f;
			if (allocHeight != 0f) {
				var offY = this.itemOffset;
				eachElement: for (int i = this.itemI; i < count;) {
					WidgetInstance<?> element;
					{
						var layoutContext = new LayoutContextImpl(this, context);
						builder.build(layoutContext, i);
						element = layoutContext.element;
					}

					element.updateMount(this);
					assert element.parent() != null;

					element.setAvailableSize(allocWidth, Float.MAX_VALUE);
					element.stencilLayout(context);

					float height = element.computedHeight();
					if (i == this.itemI && i + 1 == count) {
						if (offY < 0.0 && offY + height < allocHeight) {
							offY = (this.itemOffset = Math.max(0, height - allocHeight));
						}
					} else if (offY <= -height) {
						if (i + 1 != count) {
							this.itemI = ++i;
							this.itemOffset += height;
							offY += height;
							continue eachElement;
						}
					}

					if (i == this.itemI && offY > 0.0) {
						if (i > 0) {
							this.itemI = --i;
//							this.itemOffset = (offY -= height);

							{
								var layoutContext = new LayoutContextImpl(this, context);
								builder.build(layoutContext, i);
								element = layoutContext.element;
							}

							element.updateMount(this);

							element.setAvailableSize(allocWidth, Float.MAX_VALUE);
							element.stencilLayout(context);

							height = element.computedHeight();

							this.itemOffset = (offY -= height);
						} else {
							this.itemOffset = offY = 0.0;
						}
						continue eachElement;
					}
					offY += height;
					children.add(element);
					width = Math.max(width, element.computedWidth());
					if (offY >= allocHeight) {
						break;
					}

					i++;
				}
			}

			setComputedSize(width, allocHeight);

			/*
			pass: while (true) {
				float width = 0;
				int i = this.itemI;
				var offY = this.itemOffset;
				eachElement: while (i < count && offY < allocHeight) {
					var widget = builder.build(i);
					{
						var availHeight = (float)(allocHeight - offY);
						var element = context.getState(this, keyFunction.apply(i), widget);

						element.updateMount(this);

						element.setAvailableSize(allocWidth, availHeight);
						element.stencilLayout(context);

						float height = element.computedHeight();

						if (i + 1 == count) {
							// if this is the last item

//						var invertedDelta = Math.min(height - availHeight, this.itemOffset);
//						offY -= invertedDelta;
//						this.itemOffset -= invertedDelta;
//						this.itemOffset -= invertedDelta;
							var target = Math.max(offY, allocHeight - height);
							var delta = target - offY;
							offY = target;
							this.itemOffset += delta;

//						if (i_ == this.itemI) {
//							// if this is the first displayed item
//
//							if (offY < 0.0 && offY + height < allocHeight) {
//								offY = (this.itemOffset = Math.max(0, height - allocHeight));
//							}
//						}
						} else {
							// if there is an item after this one

							if (offY <= -height) {
								// if the scroll offset is such that the item is above the top bound of the viewport

								assert i == this.itemI;

								this.itemI = ++i;
								this.itemOffset = offY += height;

								// retry with the new element
								continue eachElement;
							}
						}


						if (i == this.itemI) {
							if (offY > 0.0) {
								do {
									boolean needsOffset;
									if (i == 0) {
										// clamp
										this.itemOffset = offY = 0.0;
										availHeight = allocHeight;
										needsOffset = false;
									} else {
										// bring new items into view from the top
										this.itemI = --i;

										element = context.getState(this, keyFunction.apply(i), widget);

										element.updateMount(this);

										needsOffset = true;
									}

									element.setAvailableSize(allocWidth, availHeight);
									element.stencilLayout(context);

									height = element.computedHeight();

									if (needsOffset) {
										this.itemOffset = (offY -= height);
										availHeight = (float) (allocHeight - offY);
									}
								} while (offY > 0.0);
								continue pass;
							}
						}

//					if (offY <= -height) {
//
//						if (i_ + 1 == count) {
//							// if there is an item after this one
//
//							this.itemI = i_ + 1;
//							this.itemOffset += height;
//							offY += height;
//						}
//					}

//					if (i == this.itemI && offY > 0.0) {
//						if (i > 0) {
//							this.itemI = --i;
//							this.itemOffset = offY -= height;
//						} else {
//							this.itemOffset = offY = 0.0;
//						}
//						continue eachElement;
//					}
						children.add(element);
						width = Math.max(width, element.computedWidth());
						offY += height;
					}

					i++;
				}

				setComputedSize(width, allocHeight);
				break;
			}
			 */
		}

		@Override
		protected void computeTraceLayout() {
			markNeedsScissorLayout();

			var children = this.children;

			final float x = this.allocatedX();
			final float minY = this.allocatedY();

			double y = minY + this.itemOffset;

			for (var child : children) {
				child.setPosition(x, (float)y);
				child.traceLayout();
				y += child.computedHeight();
			}
		}

		@Override
		protected void computeScissorLayout() {
			var clippedWidth = clipExtent(this.clippedX(),
					this.clippedWidth(),
					this.allocatedX(),
					this.computedWidth());
			var clippedHeight = clipExtent(this.clippedY(),
					this.clippedHeight(),
					this.allocatedY(),
					this.computedHeight());
			this.children.forEach(child -> {
				child.setClippedRect(clippedWidth.start(), clippedHeight.start(), clippedWidth.extent(), clippedHeight.extent());
				child.scissorLayout();
			});
		}

		@Override
		protected void computeMount(MountContext context) {
			context.pushScissor(this);
			context.mountList(this.children);
			context.popScissor();
		}

		@Override
		public void uSpecMount(MountContext context) {
			super.uSpecMount(context);
		}

		@Override
		public boolean mouseScrolled(double mouseX, double mouseY, double delta) {
			if (isInside(mouseX, mouseY)) {
				this.itemOffset += MouseScrollHandler.applyStandardDeltaTransform(delta);
				this.markNeedsStencilLayout();
				return true;
			}
			return false;
		}

		private static final class LayoutContextImpl implements LayoutContext {
			private final StateContext context;
			private final State state;
			private WidgetInstance<?> element;

			public LayoutContextImpl(State state, StateContext context) {
				this.state = state;
				this.context = context;
			}

			@Override
			public void layout(Key.CallSite callSite, Widget child) {
				this.element = this.context.getState(this.state, callSite, child);
			}
		}
	}
}
