package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.StatefulWidgetInstance;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.invoke.MethodType;

import static java.lang.invoke.MethodType.methodType;

/**
 * A widget that calls the provided {@code builder} once per frame. This is the preferred way to create widgets that
 * animate or for which diffing would be too frequent and expensive.
 * <p>
 * The {@code supplier} is tracked. It is called again every frame.
 */
public record FrameBuilderWidget(Builder builder) implements Widget {
	@Override
	public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	@FunctionalInterface
	public interface Builder {
		MethodType MT_build = methodType(Widget.class, long.class);

		/**
		 * @param time the current time as returned by {@link System#nanoTime()}.
		 */
		Widget build(long time);
	}

	private static final class State extends StatefulWidgetInstance<FrameBuilderWidget> {
		private Builder builder;

		private @Nullable ListenerDisposer updater;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		protected @NotNull Widget build() {
			return builder.build(System.nanoTime());
		}

		@Override
		public void configure(StateContext stateContext, FrameBuilderWidget widget, boolean init) {
			builder = widget.builder;
		}

		@Override
		protected void onMount() {
			super.onMount();

			if (updater == null) {
				updater = guiContext().preUpdate.listenAndInvoke(this::markNeedsRebuild);
			}
		}

		@Override
		public void onUnmount() {
			if (updater != null) {
				updater.close();
				updater = null;
			}
			super.onUnmount();
		}
	}
}
