package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.StatefulWidgetInstance;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;

/**
 * @param future  The future to listen to. Tracked by identity.
 * @param builder The builder function. Tracked by identity, low overhead of change.
 */
public record FutureBuilder<T>(CompletableFuture<T> future, Builder<T> builder) implements Widget {
	public FutureBuilder(CompletableFuture<T> fut, Function<T, @NotNull Widget> successBuilder, Function<Throwable, @NotNull Widget> failureBuilder) {
		this(fut, successBuilder, failureBuilder, Widgets.EMPTY);
	}

	public FutureBuilder(CompletableFuture<T> fut, Function<T, @NotNull Widget> successBuilder, Function<Throwable, @NotNull Widget> failureBuilder
			, @NotNull Widget childWhileLoading) {
		this(fut, new SplitBuilder<>(successBuilder, failureBuilder, childWhileLoading));
	}

	@Override
	public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
		return new State<>(args);
	}

	private static final class State<T> extends StatefulWidgetInstance<FutureBuilder<T>> {
		private static final AltResult NULL_RESULT = new AltResult(null);

		private final AtomicInteger generation = new AtomicInteger();
		private CompletableFuture<T> currentFuture;
		private Builder<T> builder;
		private @Nullable Object result;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		protected Widget build() {
			var result = this.result;
			Throwable ex;
			boolean c = result != null;
			if (!c) {
				ex = null;
			} else if (result instanceof State.AltResult failed) {
				ex = failed.ex();
				result = null;
			} else {
				ex = null;
			}
			return Objects.requireNonNull(builder.accept((T) result, ex, c), "builder returned null");
		}

		private void setFuture(CompletableFuture<T> future) {
			int expectedGeneration = generation.addAndGet(1);
			currentFuture = future;
			result = null;
			future.whenComplete((t, e) -> {
				// abort if generations do not match, meaning that the future was replaced.
				if (generation.get() != expectedGeneration) return;
				//noinspection UnreachableCode,ConstantValue
				result = e == null ? t == null ? NULL_RESULT : t : new AltResult(e);
				markNeedsRebuild();
			});
		}

		@Override
		public void configure(StateContext stateContext, FutureBuilder<T> widget, boolean init) {
			boolean dirty = init;
			if (!widget.builder.equals(builder)) {
				dirty = true;
				builder = widget.builder;
			}
			CompletableFuture<T> fut = widget.future;
			if (fut != currentFuture) {
				dirty = true;
				setFuture(fut);
			}
			if (dirty) markNeedsRebuild();
		}

		private record AltResult(@Nullable Throwable ex) {}
	}

	@FunctionalInterface
	public interface Builder<T> {
		@NotNull Widget accept(T value, @Nullable Throwable throwable, boolean complete);
	}

	private record SplitBuilder<T>(Function<T, @NotNull Widget> successBuilder,
									 Function<Throwable, @NotNull Widget> failureBuilder,
									 @NotNull Widget loadingWidget) implements Builder<T> {
		public SplitBuilder {
			Objects.requireNonNull(successBuilder);
			Objects.requireNonNull(failureBuilder);
			Objects.requireNonNull(loadingWidget);
		}

		@Override
		public @NotNull Widget accept(T t, @Nullable Throwable e, boolean c) {
			if (!c) return loadingWidget;
			if (e != null) return Objects.requireNonNull(failureBuilder.apply(e), "failureBuilder returned null");
			return Objects.requireNonNull(successBuilder.apply(t), "successBuilder returned null");
		}
	}
}
