package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.SingleChildWidgetInstance;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.time.Duration;
import java.util.List;

import static dev.pfaff.altargui.widgets.MultiChildWidget.stack;

/**
 * Inspired by the HTML <a href="https://en.wikipedia.org/wiki/Marquee_element">marquee element</a>. In contrast to
 * that element, however, this element will loop the element while it is still exiting the screen.
 * <p>
 * As a result of this, if the widget has managed state, there will be two independent copies of the state.
 */
public record MarqueeWidget(Widget content, Duration cycleDuration) implements Widget {
	@Override
	public @NonNull WidgetInstance<MarqueeWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends SingleChildWidgetInstance<MarqueeWidget> {
		@Nullable
		private Widget content;
		@Nullable
		private Widget inner;
		@Nullable
		private Duration cycleDuration;

		private long cycleDurationNanos;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, MarqueeWidget widget, boolean init) {
			var contentDirty = !widget.content.equals(this.content);
			Widget inner;
			if (contentDirty) {
				var content = this.content = widget.content;
				inner = this.inner = stack(List.of(
					new CustomTransformBuilder(content,
						(transform, _, _, allocatedWidth, _) -> {
							transform.x -= allocatedWidth;
						}),
					content
				));
			} else {
				assert this.inner != null;
				inner = this.inner;
			}
			if (!widget.cycleDuration.equals(this.cycleDuration)) {
				this.cycleDurationNanos = (this.cycleDuration = widget.cycleDuration).toNanos();
			}
			if (contentDirty) {
				setChild(stateContext, new FrameBuilderWidget(time -> {
					long cycleDuration = this.cycleDurationNanos;
					var progress = (float) ((time % cycleDuration) / (double) cycleDuration);
					return new CustomTransformBuilder(inner, (transform, _, _, allocatedWidth, _) -> {
						transform.x = progress * allocatedWidth;
					});
				}));
			}
		}
	}
}
