package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.StatefulWidgetInstance;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.time.Duration;

/**
 * A widget that facilitates creating entrance transitions.
 *
 * @param onlyFirstMount whether to restart the transition upon remount. Defaults to {@code true}.
 */
public record MountTransitionWidget(Builder builder, Duration duration, boolean onlyFirstMount) implements Widget {
	public MountTransitionWidget(Builder builder, Duration duration) {
		this(builder, duration, false);
	}

	@Override
	public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	@FunctionalInterface
	public interface Builder {
		Widget build(Duration sinceMounted);
	}

	private static final class State extends StatefulWidgetInstance<MountTransitionWidget> {
		private long mountedTime;

		private Builder builder;

		private boolean onlyFirstMount;

		private Duration duration;

		private @Nullable ListenerDisposer updater;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		private void resetTime() {
			mountedTime = System.nanoTime();
		}

		private void startUpdating() {
			if (updater != null) {
				throw new IllegalStateException();
			}
			updater = guiContext().preUpdate.listenAndInvoke(this::markNeedsRebuild);
		}

		@Override
		protected @NotNull Widget build() {
			var elapsed = Duration.ofNanos(System.nanoTime() - mountedTime);
			if (updater != null && elapsed.compareTo(duration) >= 0) {
				updater.close();
				updater = null;
			}
			return builder.build(elapsed);
		}

		@Override
		public void configure(StateContext stateContext, MountTransitionWidget widget, boolean init) {
			if (builder != widget.builder) {
				builder = widget.builder;
				markNeedsRebuild();
			}
			if (duration != widget.duration) {
				duration = widget.duration;
				if (updater == null && !init) {
					startUpdating();
				}
			}
			onlyFirstMount = widget.onlyFirstMount;
			if (init) {
				resetTime();
				startUpdating();
			}
		}

		@Override
		protected void onMount() {
			super.onMount();
			if (!onlyFirstMount) {
				resetTime();
				if (updater == null) startUpdating();
			}
		}

		@Override
		protected void onUnmount() {
			if (updater != null) {
				updater.close();
				updater = null;
			}
			super.onUnmount();
		}
	}
}
