package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.handlers.MouseScrollHandler;
import dev.pfaff.altargui.inspector.DebugLayer;
import dev.pfaff.altargui.internal.util.StencilLayoutContext;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;

import static dev.pfaff.altargui.util.ClipUtil.clipExtent;

public record ScrollWidget(Widget child) implements SingleChildWidget {
	public static final DebugLayer DEBUG_LAYER_PRE_UNSCISSORED = new DebugLayer(ScrollWidget.class, "pre_unscissored");
	public static final DebugLayer DEBUG_LAYER_PRE_SCISSORED = new DebugLayer(ScrollWidget.class, "pre_scissored");
	public static final DebugLayer DEBUG_LAYER_POST_UNSCISSORED = new DebugLayer(ScrollWidget.class, "post_unscissored");

	public ScrollWidget {
		Objects.requireNonNull(child);
	}

	@Override
	public @NotNull AbstractState<ScrollWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends AbstractState<ScrollWidget> implements MouseScrollHandler {
		/**
		 * The scroll offset.
		 */
		private double offset;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		protected void computeStencilLayout(StateContext context) {
			final var allocWidth = this.allocatedWidth();
			final var allocHeight = this.allocatedHeight();

			var childElement = getChildElement();
			StencilLayoutContext.updateMountAndLayout(context, this, childElement, (e, ctx) -> {
				e.setAvailableSize(allocWidth, allocHeight);
				e.stencilLayout(ctx);
			});

			float height = childElement.computedHeight();
			if (offset < 0.0 && height < allocHeight) {
				offset = Math.min(0, allocHeight - height);
			}

			setComputedSize(childElement.computedWidth(), Math.min(height, allocHeight));
		}

		@Override
		protected void computeTraceLayout() {
			var y = (float) (this.allocatedY() + offset);

			var childElement = getChildElement();
			childElement.setPosition(allocatedX(), y);
			childElement.traceLayout();
		}

		@Override
		protected void computeScissorLayout() {
			var clippedWidth = clipExtent(this.clippedX(),
					this.clippedWidth(),
					this.allocatedX(),
					this.computedWidth());
			var clippedHeight = clipExtent(this.clippedY(),
					this.clippedHeight(),
					this.allocatedY(),
					this.computedHeight());
			var childElement = this.getChildElement();
			childElement.setClippedRect(clippedWidth.start(), clippedHeight.start(), clippedWidth.extent(), clippedHeight.extent());
			childElement.scissorLayout();
		}

		@Override
		protected void computeMount(MountContext context) {
			context.pushScissor(this);
			super.computeMount(context);
			context.popScissor();
		}

		@Override
		public void uSpecMount(MountContext context) {
			super.uSpecMount(context);
		}

		@Override
		public boolean mouseScrolled(double mouseX, double mouseY, double delta) {
			if (isInside(mouseX, mouseY)) {
				offset = Math.min(0, offset + MouseScrollHandler.applyStandardDeltaTransform(delta));
				markNeedsLayout();
				return true;
			}
			return false;
		}
	}
}
