package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.internal.util.StencilLayoutContext;
import org.jetbrains.annotations.NotNull;

public interface SizedWidget extends SingleChildWidget {
	public record Absolute(float width, float height, Widget child) implements SizedWidget {
		@Override
		public @NotNull AbstractState<Absolute> instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends AbstractState<Absolute> {
			private float width;
			private float height;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, Absolute widget, boolean init) {
				super.configure(stateContext, widget, init);
				if (width != widget.width || height != widget.height) {
					width = widget.width;
					height = widget.height;
					markNeedsLayout();
				}
			}

			@Override
			protected void computeStencilLayout(StateContext context) {
				var width = this.width;
				var height = this.height;
				StencilLayoutContext.updateMountAndLayout(context, this, getChildElement(), (element, ctx) -> {
					element.setAvailableSize(width, height);
					element.stencilLayout(ctx);
				});
				setComputedSize(width, height);
			}

			@Override
			protected void computeTraceLayout() {
				var e = getChildElement();
				e.copyPositionFrom(this);
				e.traceLayout();
			}
		}
	}

	public record AbsoluteWidth(float width, Widget child) implements SizedWidget {
		@Override
		public @NotNull AbstractState<AbsoluteWidth> instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends AbstractState<AbsoluteWidth> {
			private float width;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, AbsoluteWidth widget, boolean init) {
				super.configure(stateContext, widget, init);
				if (width != widget.width) {
					width = widget.width;
					markNeedsLayout();
				}
			}

			@Override
			protected void computeStencilLayout(StateContext context) {
				var width = this.width;
				StencilLayoutContext.updateMountAndLayout(context, this, getChildElement(), (e, ctx) -> {
					e.setAvailableSize(width, allocatedHeight());
					e.stencilLayout(ctx);
					setComputedSize(width, e.computedHeight());
				});
			}

			@Override
			protected void computeTraceLayout() {
				var e = getChildElement();
				e.copyPositionFrom(this);
				e.traceLayout();
			}
		}
	}

	public record AbsoluteHeight(float height, Widget child) implements SizedWidget {
		@Override
		public @NotNull AbstractState<AbsoluteHeight> instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends AbstractState<AbsoluteHeight> {
			private float height;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, AbsoluteHeight widget, boolean init) {
				super.configure(stateContext, widget, init);
				if (height != widget.height) {
					height = widget.height;
					markNeedsLayout();
				}
			}

			@Override
			protected void computeStencilLayout(StateContext context) {
				var height = this.height;
				StencilLayoutContext.updateMountAndLayout(context, this, getChildElement(), (e, ctx) -> {
					e.setAvailableSize(allocatedWidth(), height);
					e.stencilLayout(ctx);
					setComputedSize(e.computedWidth(), height);
				});
			}

			@Override
			protected void computeTraceLayout() {
				var e = getChildElement();
				e.copyPositionFrom(this);
				e.traceLayout();
			}
		}
	}

	/**
	 * Expands to fill the allocated space. Parents can optionally yield space to non-expand siblings.
	 */
	public record Expand(Widget child, boolean x, boolean y) implements SizedWidget {
		public Expand {
			if (child instanceof Expand) {
				throw new IllegalArgumentException("Directly nesting Expand widgets is an error because it will not provide the desired behaviour.");
			}
		}

		@Override
		public @NotNull AbstractState<Expand> instantiate(WidgetInstantiationArgs args) {
			return new State(args);
		}

		private static final class State extends AbstractState<Expand> {
			private boolean x;
			private boolean y;

			private State(WidgetInstantiationArgs args) {
				super(args);
			}

			@Override
			public void configure(StateContext stateContext, Expand widget, boolean init) {
				super.configure(stateContext, widget, init);
				this.x = widget.x;
				this.y = widget.y;
			}

			@Override
			protected void computeStencilLayout(StateContext context) {
				StencilLayoutContext.updateMountAndLayout(context, this, getChildElement(), (element, ctx) -> {
					var width = allocatedWidth();
					var height = allocatedHeight();
					element.setAvailableSize(width, height);
					element.stencilLayout(ctx);
					setComputedSize(this.x ? width : element.computedWidth(),
							this.y ? height : element.computedHeight());
				});
			}

			@Override
			protected void computeTraceLayout() {
				var e = getChildElement();
				e.copyPositionFrom(this);
				e.traceLayout();
			}
		}
	}
}
