package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.element.LeafWidgetInstance;
import dev.pfaff.altargui.element.Painter;
import dev.pfaff.altargui.function.CharProvider;
import dev.pfaff.altargui.render.objects.TextRenderObject;
import dev.pfaff.altargui.text.Font;
import dev.pfaff.altargui.text.FontManager;
import dev.pfaff.altargui.text.StyleFlags;
import dev.pfaff.altargui.types.Color;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import dev.pfaff.altargui.util.listenable.Value;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.util.function.IntUnaryOperator;

import static dev.pfaff.altargui.render.TextRenderLayerUtil.*;
import static dev.pfaff.altargui.text.FontManager.ON_RELOAD;
import static dev.pfaff.altargui.text.StyleFlags.BOLD;
import static dev.pfaff.altargui.text.StyleFlags.isBold;

/**
 * A widget that displays a segment of uniformly-styled text.
 *
 * @param color text colorRgba (ARGB)
 * @param flags {@link StyleFlags}
 */
public record TextFragmentWidget(CharProvider characters, @NonNull Font font, int color, int flags, Value<Float> fontSize) implements TextWidget {
	/**
	 * @deprecated Use {@link #TextFragmentWidget(CharProvider, Font, Color, int, Value)} instead.
	 */
	@Deprecated
	public TextFragmentWidget {
	}

	public TextFragmentWidget(CharProvider characters, @NonNull Font font, Color color, int flags, Value<Float> fontSize) {
		this(characters, font, color.rgba(), flags, fontSize);
	}

	public TextFragmentWidget(CharProvider characters, @NonNull Font font) {
		// TODO: this is too bright. Find the correct default colorRgba value.
		this(characters, font, DEFAULT_COLOR, StyleFlags.REGULAR, TextWidget.DEFAULT_SCALED_FONT_SIZE_LISTENABLE);
	}

	public TextFragmentWidget(char character, @NonNull Font font) {
		this(CharProvider.of(character), font);
	}

	public TextFragmentWidget(String string, @NonNull Font font) {
		this(CharProvider.of(string), font);
	}

	@Override
	public TextWidget withFont(Font font) {
		return new TextFragmentWidget(characters, font, color, flags, fontSize);
	}

	@Override
	public TextWidget withFontSize(Value<Float> fontSize) {
		return new TextFragmentWidget(characters, font, color, flags, fontSize);
	}

	public TextFragmentWidget withColor(Color color) {
		return new TextFragmentWidget(characters, font, color, flags, fontSize);
	}

	@Override
	public TextFragmentWidget mapFlags(IntUnaryOperator modifier) {
		return new TextFragmentWidget(characters, font, color, modifier.applyAsInt(flags), fontSize);
	}

	public TextFragmentWidget withFlags(int flags) {
		return (TextFragmentWidget) TextWidget.super.withFlags(flags);
	}

	public TextFragmentWidget withBold(boolean bold) {
		return (TextFragmentWidget) TextWidget.super.withBold(bold);
	}

	public TextFragmentWidget withItalic(boolean italic) {
		return (TextFragmentWidget) TextWidget.super.withItalic(italic);
	}

	public TextFragmentWidget withStrike(boolean strike) {
		return (TextFragmentWidget) TextWidget.super.withStrike(strike);
	}

	public TextFragmentWidget withUnderline(boolean underline) {
		return (TextFragmentWidget) TextWidget.super.withUnderline(underline);
	}

	@Override
	public @NonNull WidgetInstance<TextFragmentWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends LeafWidgetInstance<TextFragmentWidget> implements Painter {
		private static final int LAYOUT_DEPENDENCY_FLAGS = BOLD;

		private final TextRenderObject renderObject = new TextRenderObject(0);
		private CharProvider characters;
		private @Nullable Font font;
		private int color;
		private int flags;
		private Value<Float> fontSizeListenable;
		private @Nullable ListenerDisposer fontSizeListener;
		private @Nullable ListenerDisposer fontReloadListener;
		private float fontSize = 8f;
		private boolean charsDirty = false;
		private boolean texDirty = false;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, TextFragmentWidget widget, boolean init) {
			boolean texDirty = !widget.font.equals(font);
			if (texDirty) {
				FontManager.load(widget.font);
			}
			// this is here for a reason
			texDirty |= characters == null || widget.characters.length() != characters.length();
			boolean layoutDirty = false;
			layoutDirty |= texDirty || !widget.characters.equals(characters);
			layoutDirty |= (widget.flags & LAYOUT_DEPENDENCY_FLAGS) != (flags & LAYOUT_DEPENDENCY_FLAGS);
			boolean charsDirty = layoutDirty;
			charsDirty |= widget.flags != flags;
			charsDirty |= widget.color != color;
			if (texDirty) {
				this.texDirty = true;
				this.font = widget.font;
			}
			if (charsDirty) {
				this.charsDirty = true;
				this.color = widget.color;
				this.characters = widget.characters;
				this.flags = widget.flags;
				markNeedsStencilLayout();
			}
//			if (layoutDirty) markNeedsLayout();
			if (!widget.fontSize.equals(fontSizeListenable)) {
				if (fontSizeListener != null) fontSizeListener.close();
				fontSizeListener = (fontSizeListenable = widget.fontSize).listenAndInvoke(size -> {
					fontSize = size;
					this.charsDirty = true;
					markNeedsStencilLayout();
				});
			}
		}

		@Override
		protected void onMount() {
			super.onMount();

			if (fontSizeListener == null) {
				fontSizeListener = fontSizeListenable.listenAndInvoke(size -> {
					fontSize = size;
					this.charsDirty = true;
					markNeedsStencilLayout();
				});
			}

			if (this.fontReloadListener == null) {
				this.fontReloadListener = ON_RELOAD.listenAndInvoke(() -> {
					this.texDirty = true;
					this.charsDirty = true;
					markNeedsStencilLayout();
				});
			}
		}

		@Override
		protected void onUnmount() {
			if (fontSizeListener != null) {
				fontSizeListener.close();
				fontSizeListener = null;
			}
			if (fontReloadListener != null) {
				fontReloadListener.close();
				fontReloadListener = null;
			}
			super.onUnmount();
		}

		@Override
		protected void computeStencilLayout(StateContext context) {
			// TODO: optimize
			if (charsDirty) {
				if (!font.isLoaded()) return;
				charsDirty = false;
				var flags = this.flags;
				var fontSize = this.fontSize;
				var boldAdvance = isBold(flags) ? BOLD_ADVANCE * fontSize : 0.0f;
				var regularSpacing = SPACING * fontSize;
				var spacing = regularSpacing + boldAdvance;
				var renderObject = this.renderObject;
				var color = this.color;
				var font = this.font;
				var characters = this.characters;
				var provider = new TextRenderObject.Provider() {
					public float x = 0f;

					@Override
					public void accept(TextRenderObject.Sink sink) {
						characters.provide(c -> {
							this.x -= font.getMinX(c) * fontSize;
							// FIXME: all characters are currently offset by +1 pixels on the X-axis.
							sink.accept((byte) c, color, flags, this.x, 0f, fontSize);
							this.x += font.getMaxX(c) * fontSize + spacing;
						});
					}
				};
//				Logger.log(() -> STR."Updating characters with \{characters.length()} characters from '\{characters}'");
				renderObject.updateCharacters(characters.length(), provider);
				// remove trailing spacing
				if (provider.x != 0f) provider.x -= regularSpacing;
//			Logger.log(() -> STR."TextFragmentWidget of \{characters} goes from \{initX},\{y} to \{consumer.x},\{y + fontSize}");
				setComputedSize(provider.x, fontSize);
			}
		}

		@Override
		protected void computeTraceLayout() {
			this.renderObject.updatePosition(allocatedX(), allocatedY());
		}

		private static GuiException throwNotEnoughSpaceException(boolean canOverflowX, boolean canOverflowY, float requiredWidth, float requiredHeight, float availableWidth, float availableHeight) {
			var sb = new StringBuilder("Not enough available space for text fragment: ");
			boolean overflowsX = !canOverflowX && requiredWidth > availableWidth;
			boolean overflowsY = !canOverflowY && requiredHeight > availableHeight;
			if (overflowsX) {
				sb.append("required width ")
						.append(requiredWidth)
						.append(" overflows X-axis by ")
						.append(requiredWidth - availableWidth);
			}
			if (overflowsX && overflowsY) {
				sb.append(" and ");
			}
			if (overflowsY) {
				sb.append("required height ")
						.append(requiredHeight)
						.append(" overflows Y-axis by ")
						.append(requiredHeight - availableHeight);
			}
			sb.append(".");
			return new GuiException(sb.toString());
		}

		@Override
		public void uSpecMount(MountContext context) {
			super.uSpecMount(context);
		}

		@Override
		public void paint(long time, Canvas canvas) {
			if (this.texDirty) {
				this.texDirty = false;
				canvas.texAlloc().alloc(this.font.texture())
						.cancelOnInterrupt()
						.thenAccept(handle -> this.renderObject.updateTex(handle.u, handle.v))
						.uncatch();
			}
			canvas.upload(this.renderObject);
		}
	}
}
