package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.StatefulWidgetInstance;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import dev.pfaff.altargui.util.listenable.Value;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.function.Function;

/**
 * A widget that replaces its child using the provided {@code builder} whenever notified by listenable.
 */
public record ValueListenableBuilderWidget(Value<@NotNull Widget> listenable) implements Widget {
	public static <T> ValueListenableBuilderWidget builder(Value<T> listenable, Function<T, @NotNull Widget> builder) {
		return new ValueListenableBuilderWidget(listenable.map(builder));
	}

	@Override
	public @NotNull WidgetInstance<ValueListenableBuilderWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends StatefulWidgetInstance<ValueListenableBuilderWidget> {
		private Value<@NotNull Widget> listenable;
		private Widget child;

		private @Nullable ListenerDisposer listener;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		protected @NotNull Widget build() {
			return this.child;
		}

		@Override
		public void configure(StateContext stateContext, ValueListenableBuilderWidget widget, boolean init) {
			if (this.listenable != widget.listenable) {
				if (this.listener != null) this.listener.close();
				this.listenable = widget.listenable;
				updateListener();
			}
		}

		private void updateListener() {
			this.listener = this.listenable.listenAndInvoke(child -> {
				//noinspection ConstantValue
				if (child == null) {
					// using this.listenable here to avoid another capture.
					throw new NullPointerException(STR."\{ValueListenableBuilderWidget.class.getName()} listenable emitted a null value: \{this.listenable}");
				}
				this.child = child;
				this.markNeedsRebuild();
			});
		}

		@Override
		protected void onMount() {
			super.onMount();

			if (this.listener == null) {
				updateListener();
			}
		}

		@Override
		protected void onUnmount() {
			if (this.listener != null) {
				this.listener.close();
				this.listener = null;
			}
			super.onUnmount();
		}
	}
}
